#!/usr/bin/env python3

from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser, FileType, Namespace
from dataclasses import dataclass
from itertools import chain
from typing import Any, Generator, Iterable, Optional

import torch
from tabulate import tabulate, tabulate_formats
from tqdm import tqdm
from transformers import (
    AutoModelForSeq2SeqLM,
    AutoTokenizer,
    M2M100ForConditionalGeneration,
    set_seed,
)
from transformers.generation.utils import GenerateOutput, GenerationMixin

from mbrs import timer


def buffer_lines(input_stream: Iterable[str], buffer_size: int = 64):
    buf: list[str] = []
    for i, line in enumerate(tqdm(input_stream)):
        buf.append(line.strip())
        if (i + 1) % buffer_size == 0:
            yield buf
            buf = []
    if len(buf) > 0:
        yield buf


def get_argparser() -> ArgumentParser:
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    # fmt: off
    parser.add_argument("input", nargs="?", default="-",
                        type=FileType("r", encoding="utf-8"),
                        help="Input file. If not specified, read from stdin.")
    parser.add_argument("--output", "-o", default="-", type=FileType("w"),
                        help="Output file.")
    parser.add_argument("--lprobs", default=None, type=FileType("w"),
                        help="Reference log-probabilities file. "
                        "This option is useful for the model-based estimation.")
    parser.add_argument("--length_normalized_lprobs", default=None, type=FileType("w"),
                        help="Length-normalized reference log-probabilities file. "
                        "This option is useful for the model-based estimation.")
    parser.add_argument("--model", "-m", type=str, default="facebook/m2m100_418M",
                        help="Model name or path.")
    parser.add_argument("--num_candidates", "-n", type=int, default=1,
                        help="Number of candidates to be returned.")
    parser.add_argument("--sampling", "-s", type=str, default="",
                        choices=["eps"],
                        help="Sampling method.")
    parser.add_argument("--beam_size", type=int, default=5,
                        help="Beam size.")
    parser.add_argument("--epsilon", "--eps", "-e", type=float, default=0.02,
                        help="Cutoff parameter for epsilon sampling.")
    parser.add_argument("--lang_pair", "-l", type=str, default="en-de",
                        help="Language name pair. Some models like M2M100 uses this information.")
    parser.add_argument("--max_length", type=int, default=1024,
                        help="Maximum length of an output sentence.")
    parser.add_argument("--min_length", type=int, default=1,
                        help="Minimum length of an output sentence.")
    parser.add_argument("--length_penalty", type=float, default=None,
                        help="Length penalty.")
    parser.add_argument("--batch_size", "-b", type=int, default=8,
                        help="Batch size.")
    parser.add_argument("--sampling_size", type=int, default=8,
                        help="Sampling size in a single inference. "
                        "The model generates this number of samples at a time "
                        "until the total number of samples reaches `--num_candidates`.")
    parser.add_argument("--fp16", action="store_true",
                        help="Use float16.")
    parser.add_argument("--bf16", action="store_true",
                        help="Use bfloat16.")
    parser.add_argument("--cpu", action="store_true",
                        help="Force to use CPU.")
    parser.add_argument("--seed", type=int, default=0,
                        help="Random number seed.")
    parser.add_argument("--quiet", "-q", action="store_true",
                        help="No report statistics.")
    parser.add_argument("--report", default="-", type=FileType("w"),
                        help="Report file.")
    parser.add_argument("--report_format", type=str, default="rounded_outline",
                        choices=tabulate_formats,
                        help="Report runtime statistics.")
    parser.add_argument("--width", "-w", type=int, default=1,
                        help="Number of digits for values of float point.")
    # fmt: on
    return parser


@dataclass
class Sample:
    """A sample generated by a model."""

    text: str
    lprob: Optional[float] = None
    length_normalized_lprob: Optional[float] = None


def memory_efficient_compute_transition_scores(
    model: GenerationMixin,
    sequences: torch.Tensor,
    scores: tuple[torch.Tensor, ...],
    beam_indices: Optional[torch.Tensor] = None,
    normalize_logits: bool = False,
) -> torch.Tensor:
    """
    Computes the transition scores of sequences given the generation scores (and beam indices, if beam search was
    used). This is a convenient method to quicky obtain the scores of the selected tokens at generation time.

    This is another implementation for memory efficiency.

    Parameters:
        model (`GenerationMixin`): Generation model.
        sequences (`torch.LongTensor`):
            The generated sequences. The second dimension (sequence_length) is either equal to `max_length` or
            shorter if all batches finished early due to the `eos_token_id`.
        scores (`tuple(torch.FloatTensor)`):
            Transition scores for each vocabulary token at each generation step. Beam transition scores consisting
            of log probabilities of tokens conditioned on log softmax of previously generated tokens in this beam.
            Tuple of `torch.FloatTensor` with up to `max_new_tokens` elements (one element for each generated token),
            with each tensor of shape `(batch_size*num_beams, config.vocab_size)`.
        beam_indices (`torch.LongTensor`, *optional*):
            Beam indices of generated token id at each generation step. `torch.LongTensor` of shape
            `(batch_size*num_return_sequences, sequence_length)`. Only required if a `num_beams>1` at
            generate-time.
        normalize_logits (`bool`, *optional*, defaults to `False`):
            Whether to normalize the logits (which, for legacy reasons, may be unnormalized).

    Return:
        `torch.Tensor`: A `torch.Tensor` of shape `(batch_size*num_return_sequences, sequence_length)` containing
            the transition scores (logits)

    Examples:

    ```python
    >>> from transformers import GPT2Tokenizer, AutoModelForCausalLM
    >>> import numpy as np

    >>> tokenizer = GPT2Tokenizer.from_pretrained("gpt2")
    >>> model = AutoModelForCausalLM.from_pretrained("openai-community/gpt2")
    >>> tokenizer.pad_token_id = tokenizer.eos_token_id
    >>> inputs = tokenizer(["Today is"], return_tensors="pt")

    >>> # Example 1: Print the scores for each token generated with Greedy Search
    >>> outputs = model.generate(**inputs, max_new_tokens=5, return_dict_in_generate=True, output_scores=True)
    >>> transition_scores = memory_efficient_compute_transition_scores(
    ...     model, outputs.sequences, outputs.scores, normalize_logits=True
    ... )
    >>> # input_length is the length of the input prompt for decoder-only models, like the GPT family, and 1 for
    >>> # encoder-decoder models, like BART or T5.
    >>> input_length = 1 if model.config.is_encoder_decoder else inputs.input_ids.shape[1]
    >>> generated_tokens = outputs.sequences[:, input_length:]
    >>> for tok, score in zip(generated_tokens[0], transition_scores[0]):
    ...     # | token | token string | log probability | probability
    ...     print(f"| {tok:5d} | {tokenizer.decode(tok):8s} | {score.numpy():.3f} | {np.exp(score.numpy()):.2%}")
    |   262 |  the     | -1.414 | 24.33%
    |  1110 |  day     | -2.609 | 7.36%
    |   618 |  when    | -2.010 | 13.40%
    |   356 |  we      | -1.859 | 15.58%
    |   460 |  can     | -2.508 | 8.14%

    >>> # Example 2: Reconstruct the sequence scores from Beam Search
    >>> outputs = model.generate(
    ...     **inputs,
    ...     max_new_tokens=5,
    ...     num_beams=4,
    ...     num_return_sequences=4,
    ...     return_dict_in_generate=True,
    ...     output_scores=True,
    ... )
    >>> transition_scores = memory_efficient_compute_transition_scores(
    ...     model, outputs.sequences, outputs.scores, outputs.beam_indices, normalize_logits=False
    ... )
    >>> # If you sum the generated tokens' scores and apply the length penalty, you'll get the sequence scores.
    >>> # Tip 1: recomputing the scores is only guaranteed to match with `normalize_logits=False`. Depending on the
    >>> # use case, you might want to recompute it with `normalize_logits=True`.
    >>> # Tip 2: the output length does NOT include the input length
    >>> output_length = np.sum(transition_scores.numpy() < 0, axis=1)
    >>> length_penalty = model.generation_config.length_penalty
    >>> reconstructed_scores = transition_scores.sum(axis=1) / (output_length**length_penalty)
    >>> print(np.allclose(outputs.sequences_scores, reconstructed_scores))
    True
    ```"""
    # 1. In absence of `beam_indices`, we can assume that we come from e.g. greedy search, which is equivalent
    # to a beam search approach were the first (and only) beam is always selected
    if beam_indices is None:
        beam_indices = torch.arange(scores[0].shape[0]).view(-1, 1).to(sequences.device)
        beam_indices = beam_indices.expand(-1, len(scores))

    # 2. Optionally normalize the logits (across the vocab dimension)
    if normalize_logits:
        scores = tuple(s.log_softmax(dim=1) for s in scores)

    # 3. cut beam_indices to longest beam length
    beam_indices_mask = beam_indices < 0
    max_beam_length = (1 - beam_indices_mask.long()).sum(-1).max()
    beam_indices = beam_indices.clone()[:, :max_beam_length]
    beam_indices_mask = beam_indices_mask[:, :max_beam_length]

    # 4. Set indices of beams that finished early to 0; such indices will be masked correctly afterwards
    beam_indices[beam_indices_mask] = 0

    # 5. multiply beam_indices with vocab size to gather correctly from scores
    beam_sequence_indices = beam_indices * model.config.vocab_size

    # 6. Define which indices contributed to scores
    cut_idx = sequences.shape[-1] - max_beam_length
    indices = sequences[:, cut_idx:] + beam_sequence_indices

    # 7. Compute scores
    transition_scores = torch.zeros(
        indices.size(), dtype=torch.float32, device=scores[0].device
    )
    for step, step_scores in enumerate(scores):
        transition_scores[:, step] = step_scores.view(-1).gather(0, indices[:, step])

    # 8. Mask out transition_scores of beams that stopped early
    transition_scores[beam_indices_mask] = 0

    return transition_scores


def main(args: Namespace) -> None:
    set_seed(args.seed)

    src_lang, tgt_lang = tuple(args.lang_pair.split("-"))
    tokenizer = AutoTokenizer.from_pretrained(args.model)
    model = AutoModelForSeq2SeqLM.from_pretrained(args.model)

    model.eval()
    for param in model.parameters():
        param.requires_grad = False
    if torch.cuda.is_available() and not args.cpu:
        if args.fp16:
            model.half()
        elif args.bf16:
            model.bfloat16()
        model.cuda()

    generation_kwargs = {
        "max_length": args.max_length,
        "min_length": args.min_length,
        "return_dict_in_generate": True,
    }
    length_penalty = getattr(model.generation_config, "length_penalty", 1.0)
    if args.length_penalty is not None:
        generation_kwargs["length_penalty"] = args.length_penalty
        length_penalty = args.length_penalty

    if args.lprobs is not None or args.length_normalized_lprobs is not None:
        generation_kwargs["output_scores"] = True

    if isinstance(model, M2M100ForConditionalGeneration):
        tokenizer.src_lang = src_lang
        generation_kwargs["forced_bos_token_id"] = tokenizer.get_lang_id(tgt_lang)

    if args.sampling == "eps":
        generation_kwargs["do_sample"] = True
        generation_kwargs["epsilon_cutoff"] = args.epsilon
        generation_kwargs["num_beams"] = 1
        generation_kwargs["early_stopping"] = False
    else:
        generation_kwargs["num_beams"] = max(args.beam_size, args.num_candidates)

    def decode(
        inputs: list[str], num_candidates: int, generation_kwargs: dict[str, Any]
    ) -> list[Sample]:
        model_inputs = tokenizer(inputs, return_tensors="pt", padding=True).to(
            device=model.device
        )
        with timer.measure("generate"):
            model_outputs: GenerateOutput = model.generate(
                **model_inputs, **generation_kwargs, num_return_sequences=num_candidates
            )

        if model_outputs.scores is None:
            return [
                Sample(s)
                for s in tokenizer.batch_decode(
                    model_outputs.sequences, skip_special_tokens=True
                )
            ]

        sequences = model_outputs.sequences
        scores = model_outputs.scores
        max_length = len(scores)
        if hasattr(tokenizer, "pad_token_id"):
            pad_token_id = tokenizer.pad_token_id
            for s in scores:
                s[:, pad_token_id] = float("-inf")
            scores = tuple(s.log_softmax(dim=1) for s in scores)
            for s in scores:
                s[:, pad_token_id] = 0.0
            sequence_lengths = max_length - (sequences.eq(pad_token_id)).sum(dim=-1)
        else:
            scores = tuple(s.log_softmax(dim=1) for s in scores)
            sequence_lengths = torch.full(
                (sequences.size(0),), fill_value=max_length, device=sequences.device
            )

        transition_scores = memory_efficient_compute_transition_scores(
            model,
            sequences,
            scores,
            beam_indices=model_outputs.get("beam_indices", None),
        )
        lprobs = transition_scores.sum(dim=-1)
        length_normalized_lprobs = lprobs / (sequence_lengths**length_penalty)
        return [
            Sample(text, lprob=lprob, length_normalized_lprob=length_normalized_lprob)
            for text, lprob, length_normalized_lprob in zip(
                tokenizer.batch_decode(sequences, skip_special_tokens=True),
                lprobs.cpu().tolist(),
                length_normalized_lprobs.cpu().tolist(),
            )
        ]

    def generate(inputs: list[str]) -> Generator[Sample, None, None]:
        if (
            not generation_kwargs.get("do_sample", False)
            or generation_kwargs.get("num_beams", 1) != 1
        ):
            yield from decode(inputs, args.num_candidates, generation_kwargs)
        else:
            samples: list[list[Sample]] = [[] for _ in range(args.batch_size)]
            for n in range(0, args.num_candidates, args.sampling_size):
                sampling_size = min(args.sampling_size, args.num_candidates - n)
                shards = decode(inputs, sampling_size, generation_kwargs)
                for i in range(args.batch_size):
                    samples[i] += shards[i * sampling_size : (i + 1) * sampling_size]
            yield from chain.from_iterable(samples)

    num_sentences = 0
    for lines in buffer_lines(args.input, buffer_size=args.batch_size):
        for sample in generate(lines):
            print(sample.text.strip(), file=args.output)
            if sample.lprob is not None and args.lprobs is not None:
                print(str(sample.lprob), file=args.lprobs)
            if (
                sample.length_normalized_lprob is not None
                and args.length_normalized_lprobs is not None
            ):
                print(
                    str(sample.length_normalized_lprob),
                    file=args.length_normalized_lprobs,
                )
            num_sentences += 1

    if not args.quiet:
        statistics = timer.aggregate().result(num_sentences)
        table = tabulate(
            statistics,
            headers="keys",
            tablefmt=args.report_format,
            floatfmt=f".{args.width}f",
        )
        print(table, file=args.report)


def cli_main():
    args = get_argparser().parse_args()
    main(args)


if __name__ == "__main__":
    cli_main()



import torch

from .comet import MetricCOMET

SOURCE = "これはテストです"
HYPOTHESES = [
    "this is a test",
    "another test",
    "this is a fest",
    "Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;",
]
REFERENCES = [
    "ref",
    "this is a test",
    "producţia de zahăr brut se exprimă în zahăr alb;",
]
SCORES = torch.Tensor(
    [
        [0.54616, 0.99257, 0.40706],
        [0.59092, 0.81587, 0.38636],
        [0.50268, 0.75507, 0.38999],
        [0.40692, 0.37781, 0.78060],
    ]
)
EXPECTED_SCORES_AGGREGATED = torch.Tensor([0.85933, 0.75443, 0.68586, 0.62650])


class TestMetricCOMET:
    def test_score(self, metric_comet: MetricCOMET):
        for i, hyp in enumerate(HYPOTHESES):
            for j, ref in enumerate(REFERENCES):
                assert torch.isclose(
                    SCORES[i, j],
                    torch.tensor(metric_comet.score(hyp, ref, SOURCE)),
                    atol=0.0005 / 100,
                )

    def test_expected_scores(self, metric_comet: MetricCOMET):
        expected_scores = metric_comet.expected_scores(HYPOTHESES, REFERENCES, SOURCE)
        torch.testing.assert_close(
            expected_scores,
            SCORES.mean(dim=1).to(metric_comet.device),
            atol=0.0005 / 100,
            rtol=1e-6,
        )

    def test_corpus_score(self, metric_comet: MetricCOMET):
        hyps = [
            "this is a test",
            "another test",
            "this is a fest",
            "Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;",
        ]
        refs = [
            "this is a test",
            "ref",
            "this is a test",
            "producţia de zahăr brut se exprimă în zahăr alb;",
        ]
        assert torch.isclose(
            torch.tensor(metric_comet.corpus_score(hyps, refs, [SOURCE] * len(hyps))),
            torch.tensor(0.77979),
        )

    def test_expected_scores_reference_aggregation(self, metric_comet: MetricCOMET):
        expected_scores = metric_comet.expected_scores_reference_aggregation(
            HYPOTHESES, REFERENCES, SOURCE
        )
        torch.testing.assert_close(
            expected_scores,
            EXPECTED_SCORES_AGGREGATED.to(metric_comet.device),
            atol=0.0005 / 100,
            rtol=1e-6,
        )



import multiprocessing
import sys

import torch

from .chrf import MetricChrF

HYPOTHESES = [
    "this is a test",
    "another test",
    "this is a fest",
    "Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;",
]
REFERENCES = [
    "ref",
    "this is a test",
    "producţia de zahăr brut se exprimă în zahăr alb;",
]
SCORES = torch.Tensor(
    [
        [7.246, 100.0, 3.411],
        [14.493, 23.060, 4.386],
        [14.493, 60.628, 3.411],
        [6.849, 7.224, 46.161],
    ],
)
EXPECTED_SCORES_AGGREGATED = torch.Tensor([23.08157, 10.29285, 16.35524, 29.16298])


class TestMetricChrF:
    def test_score(self):
        metric = MetricChrF(MetricChrF.Config())
        for i, hyp in enumerate(HYPOTHESES):
            for j, ref in enumerate(REFERENCES):
                assert torch.isclose(
                    SCORES[i, j], torch.tensor(metric.score(hyp, ref)), atol=0.0005
                )

    def test_expected_scores(self):
        metric = MetricChrF(MetricChrF.Config())
        expected_scores = metric.expected_scores(HYPOTHESES, REFERENCES)
        torch.testing.assert_close(
            expected_scores,
            SCORES.mean(dim=1),
            atol=0.0005,
            rtol=1e-4,
        )
        if sys.platform == "linux":
            default_method = multiprocessing.get_start_method()
            multiprocessing.set_start_method("spawn", force=True)
            expected_scores = metric.expected_scores(HYPOTHESES, REFERENCES)
            torch.testing.assert_close(
                expected_scores, SCORES.mean(dim=1), atol=0.0005, rtol=1e-4
            )
            multiprocessing.set_start_method(default_method, force=True)

    def test_scores(self):
        hyps = [
            "this is a test",
            "another test",
            "this is a fest",
            "Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;",
        ]
        refs = [
            "this is a test",
            "ref",
            "this is a test",
            "producţia de zahăr brut se exprimă în zahăr alb;",
        ]

        metric = MetricChrF(MetricChrF.Config())
        torch.testing.assert_close(
            metric.scores(hyps, refs),
            torch.tensor([metric.score(h, r) for h, r in zip(hyps, refs)]),
        )

    def test_corpus_score(self):
        hyps = [
            "this is a test",
            "another test",
            "this is a fest",
            "Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;",
        ]
        refs = [
            "this is a test",
            "ref",
            "this is a test",
            "producţia de zahăr brut se exprimă în zahăr alb;",
        ]

        metric = MetricChrF(MetricChrF.Config())
        assert torch.isclose(
            torch.tensor(metric.corpus_score(hyps, refs)), torch.tensor(53.90979)
        )

    def test_expected_scores_reference_aggregation(self):
        metric = MetricChrF(MetricChrF.Config())
        expected_scores = metric.expected_scores_reference_aggregation(
            HYPOTHESES, REFERENCES
        )
        torch.testing.assert_close(
            expected_scores, EXPECTED_SCORES_AGGREGATED, atol=0.0005, rtol=1e-4
        )

        expected_scores = metric.expected_scores_reference_aggregation(
            HYPOTHESES,
            REFERENCES,
            reference_lprobs=torch.Tensor([-2.000]).repeat(len(REFERENCES)),
        )
        torch.testing.assert_close(
            expected_scores, EXPECTED_SCORES_AGGREGATED, atol=0.0005, rtol=1e-4
        )

    def test_expected_scores_reference_aggregation_empty_inputs(self):
        metric = MetricChrF(MetricChrF.Config())
        hyps = ["thank you", ""]
        refs = ["thank you so much", "", "thank you.", "thank you", ""]
        expected_scores = metric.expected_scores_reference_aggregation(hyps, refs)
        torch.testing.assert_close(
            expected_scores, torch.Tensor([68.2679, 0.0]), atol=0.0005, rtol=1e-4
        )

        expected_scores = metric.expected_scores_reference_aggregation(
            hyps, refs, reference_lprobs=torch.Tensor([-2.000]).repeat(len(refs))
        )
        torch.testing.assert_close(
            expected_scores, torch.Tensor([68.2679, 0.0]), atol=0.0005, rtol=1e-4
        )



from typing import Optional

import torch
from torch import Tensor


def expectation(matrix: Tensor, lprobs: Optional[Tensor] = None) -> Tensor:
    """Compute expectation values for each row.

    Args:
        matrix (Tensor): Input matrix of shape `(H, R)`.
        lprobs (Tensor, optional): Log-probabilities for each column of shape `(R,)`.

    Returns:
        Tensor: Expected values for each row of shape `(H,)`.
    """
    if lprobs is None:
        return matrix.mean(dim=-1)
    else:
        if list(lprobs.shape) != list(matrix.shape)[1:]:
            raise ValueError(
                f"`weights` must have {list(matrix.shape)[1:]} elements, but got {list(lprobs.shape)}"
            )

        return (
            (
                matrix.float()
                * lprobs.to(matrix).softmax(dim=-1, dtype=torch.float32)[None, :]
            )
            .sum(dim=-1)
            .to(matrix)
        )



from __future__ import annotations

import abc
from dataclasses import dataclass, fields, make_dataclass
from typing import Any, Optional

from torch import Tensor

from mbrs.metrics.base import Metric, MetricBase, MetricReferenceless
from mbrs.selectors import Selector, SelectorNbest


class DecoderBase(abc.ABC):
    """Decoder base class."""

    def __init__(
        self,
        cfg: DecoderBase.Config,
        metric: MetricBase,
        selector: Selector = SelectorNbest(SelectorNbest.Config()),
    ) -> None:
        self.cfg = cfg
        self.metric = metric
        self.selector = selector

    @property
    def maximize(self) -> bool:
        """Return `True` when maximizing the objective score."""
        return self.metric.HIGHER_IS_BETTER

    @dataclass
    class Config:
        """Configuration for the decoder."""

    @dataclass
    class Output:
        """
        - idx (list[int]): Index numbers of the n-best hypotheses.
        - sentence (list[str]): Sentences of the n-best hypotheses.
        - score (list[float]): Scores of the n-best hypotheses.
        """

        idx: list[int]
        sentence: list[str]
        score: list[float]

        def __or__(self, other: Any):
            """Returns the union of dataclasses.

            Args:
                other (Any): An other dataclass.

            Returns:
                Output: New dataclass with the merged attributes of `self` and `other`.
            """
            new_fields = [(f.name, f.type, f) for f in fields(self)]
            new_fields += [
                (f.name, f.type, f)
                for f in fields(other)
                if f.name not in {f.name for f in fields(self)}
            ]
            new_dc_type = make_dataclass(
                "Output", fields=new_fields, bases=(type(self),)
            )
            attrs = {f.name: getattr(other, f.name) for f in fields(other)} | {
                f.name: getattr(self, f.name) for f in fields(self)
            }
            return new_dc_type(**attrs)

    def topk(self, x: Tensor, k: int = 1) -> tuple[list[float], list[int]]:
        """Return the top-k best elements and corresponding indices.

        Args:
            x (Tensor): Input 1-D array.
            k (int): Return the top-k values and indices.

        Returns:
            tuple[list[float], list[int]]
              - list[float]: The top-k values.
              - list[int]: The top-k indices.
        """
        return self.selector.topk(x, k=k, maximize=self.maximize)

    def argbest(self, x: Tensor) -> Tensor:
        """Return the index of the best element.

        Args:
            x (Tensor): Input 1-D array.

        Returns:
            Tensor: A scalar tensor of the best index.
        """
        return self.selector.argbest(x, maximize=self.maximize)

    def superior(self, a: float, b: float) -> bool:
        """Return whether the score `a` is superior to the score `b`.

        Args:
            a (float): A score.
            b (float): A score.

        Returns:
            bool: Return True when `a` is superior to `b`.
        """
        return self.selector.superior(a, b, maximize=self.maximize)

    def select(
        self,
        hypotheses: list[str],
        expected_scores: Tensor,
        nbest: int = 1,
        source: Optional[str] = None,
        *args,
        **kwargs,
    ) -> Selector.Output:
        """Select the final output list.

        Args:
            hypotheses (list[str]): Hypotheses.
            expected_scores (Tensor): The expected scores for each hypothesis.
            nbest (int): Return the n-best hypotheses based on the selection rule.
            source (str, optional): A source.
            maximize (bool): Whether maximize the scores or not.

        Returns:
            Selector.Output: Outputs.
        """
        return self.selector.select(
            hypotheses,
            expected_scores,
            nbest=nbest,
            source=source,
            maximize=self.maximize,
            *args,
            **kwargs,
        )


class DecoderReferenceBased(DecoderBase, metaclass=abc.ABCMeta):
    """Decoder base class for strategies that use references like MBR decoding."""

    metric: Metric

    @abc.abstractmethod
    def decode(
        self,
        hypotheses: list[str],
        references: list[str],
        source: Optional[str] = None,
        nbest: int = 1,
        reference_lprobs: Optional[Tensor] = None,
    ) -> DecoderReferenceBased.Output:
        """Select the n-best hypotheses based on the strategy.

        Args:
            hypotheses (list[str]): Hypotheses.
            references (list[str]): References.
            source (str, optional): A source.
            nbest (int): Return the n-best hypotheses.
            reference_lprobs (Tensor, optional): Log-probabilities for each reference sample.
              The shape must be `(len(references),)`. See `https://arxiv.org/abs/2311.05263`.

        Returns:
            Decoder.Output: The n-best hypotheses.
        """


class DecoderReferenceless(DecoderBase, metaclass=abc.ABCMeta):
    """Decoder base class for reference-free strategies."""

    metric: MetricReferenceless

    @abc.abstractmethod
    def decode(
        self, hypotheses: list[str], source: str, nbest: int = 1
    ) -> DecoderReferenceless.Output:
        """Select the n-best hypotheses based on the strategy.

        Args:
            hypotheses (list[str]): Hypotheses.
            source (str): A source.
            nbest (int): Return the n-best hypotheses.

        Returns:
            Decoder.Output: The n-best hypotheses.
        """



import pytest
import torch

from .nbest import SelectorNbest


class TestSelectorNbest:
    @pytest.mark.parametrize("nbest", [1, 3, 5])
    @pytest.mark.parametrize("maximize", [True, False])
    def test_select(self, nbest: int, maximize: bool):
        selector = SelectorNbest(SelectorNbest.Config())
        scores = torch.Tensor([0.5, 0.9, 0.8, 0.2, 0.4])
        sentences = ["a", "b", "c", "d", "e"]
        output = selector.select(sentences, scores, nbest=nbest, maximize=maximize)
        if maximize:
            expected_scores, expected_indices = scores.sort(descending=True)
        else:
            expected_scores, expected_indices = scores.sort()
        expected_sentences = [sentences[i] for i in expected_indices.tolist()]
        assert output.idx == expected_indices.tolist()[:nbest]
        assert output.sentence == expected_sentences[:nbest]
        assert torch.allclose(torch.tensor(output.score), expected_scores[:nbest])



import pytest
import torch

from mbrs.metrics import MetricCOMETkiwi
from mbrs.selectors import Selector

from .rerank import DecoderRerank

SOURCE = "これはテストです"
HYPOTHESES = [
    ["another test", "this is a test", "this is a fest"],
    ["another test", "this is a fest", "this is a test"],
    ["this is a test"],
    ["Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;"],
]

BEST_INDICES = [1, 2, 0, 0]
BEST_SENTENCES = [
    "this is a test",
    "this is a test",
    "this is a test",
    "Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;",
]
SCORES = torch.Tensor([0.86415, 0.86415, 0.86415, 0.29771])


class TestDecoderRerank:
    def test_decode(self, metric_cometkiwi: MetricCOMETkiwi):
        decoder = DecoderRerank(DecoderRerank.Config(), metric_cometkiwi)
        for i, hyps in enumerate(HYPOTHESES):
            output = decoder.decode(hyps, SOURCE, nbest=1)
            assert output.idx[0] == BEST_INDICES[i]
            assert output.sentence[0] == BEST_SENTENCES[i]
            torch.testing.assert_close(
                torch.tensor(output.score[0]),
                SCORES[i],
                atol=0.0005 / 100,
                rtol=1e-6,
            )

    @pytest.mark.parametrize("nbest", [1, 2])
    def test_decode_selector(
        self, metric_cometkiwi: MetricCOMETkiwi, nbest: int, selector: Selector
    ):
        decoder = DecoderRerank(
            DecoderRerank.Config(), metric_cometkiwi, selector=selector
        )
        for i, hyps in enumerate(HYPOTHESES):
            output = decoder.decode(hyps, SOURCE, nbest=nbest)
            assert len(output.sentence) == min(nbest, len(hyps))
            assert len(output.score) == min(nbest, len(hyps))



import multiprocessing
import sys

import torch

from .ter import MetricTER

HYPOTHESES = [
    "this is a test",
    "another test",
    "this is a fest",
    "Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;",
]
REFERENCES = [
    "ref",
    "this is a test",
    "producţia de zahăr brut se exprimă în zahăr alb;",
]
SCORES = torch.Tensor(
    [
        [400.0, 0.0, 100.0],
        [200.0, 75.0, 100.0],
        [400.0, 25.0, 100.0],
        [1300.0, 325.0, 100.0],
    ],
)


class TestMetricTER:
    def test_score(self):
        metric = MetricTER(MetricTER.Config())
        for i, hyp in enumerate(HYPOTHESES):
            for j, ref in enumerate(REFERENCES):
                assert torch.isclose(
                    SCORES[i, j], torch.tensor(metric.score(hyp, ref)), atol=0.0005
                )

    def test_expected_scores(self):
        metric = MetricTER(MetricTER.Config())
        expected_scores = metric.expected_scores(HYPOTHESES, REFERENCES)
        torch.testing.assert_close(
            expected_scores,
            SCORES.mean(dim=1),
            atol=0.0005,
            rtol=1e-4,
        )

        if sys.platform == "linux":
            default_method = multiprocessing.get_start_method()
            multiprocessing.set_start_method("spawn", force=True)
            expected_scores = metric.expected_scores(HYPOTHESES, REFERENCES)
            torch.testing.assert_close(
                expected_scores, SCORES.mean(dim=1), atol=0.0005, rtol=1e-4
            )
            multiprocessing.set_start_method(default_method, force=True)

    def test_scores(self):
        hyps = [
            "this is a test",
            "another test",
            "this is a fest",
            "Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;",
        ]
        refs = [
            "this is a test",
            "ref",
            "this is a test",
            "producţia de zahăr brut se exprimă în zahăr alb;",
        ]

        metric = MetricTER(MetricTER.Config())
        torch.testing.assert_close(
            metric.scores(hyps, refs),
            torch.tensor([metric.score(h, r) for h, r in zip(hyps, refs)]),
        )

    def test_corpus_score(self):
        hyps = [
            "this is a test",
            "another test",
            "this is a fest",
            "Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;",
        ]
        refs = [
            "this is a test",
            "ref",
            "this is a test",
            "producţia de zahăr brut se exprimă în zahăr alb;",
        ]

        metric = MetricTER(MetricTER.Config())
        assert torch.isclose(
            torch.tensor(metric.corpus_score(hyps, refs)), torch.tensor(66.66667)
        )



from __future__ import annotations

import concurrent.futures
import itertools
import math
from collections import Counter
from dataclasses import dataclass
from typing import Optional

import torch
from sacrebleu.metrics.bleu import BLEU, MAX_NGRAM_ORDER
from sacrebleu.metrics.helpers import extract_all_word_ngrams
from torch import Tensor

from mbrs import timer

from . import MetricAggregatable, register


@register("bleu")
class MetricBLEU(MetricAggregatable):
    """BLEU metric class."""

    @dataclass
    class Config(MetricAggregatable.Config):
        """BLEU metric configuration.

        - lowercase (bool): If True, lowercased BLEU is computed.
        - force (bool): Ignore data that looks already tokenized.
        - tokenize (str, optional): The tokenizer to use. If None, defaults to language-specific tokenizers with '13a' as the fallback default.
        - smooth_method (str): The smoothing method to use ('floor', 'add-k', 'exp' or 'none').
        - smooth_value (float, optional): The smoothing value for `floor` and `add-k` methods. `None` falls back to default value.
        - max_ngram_order (int): If given, it overrides the maximum n-gram order (default: 4) when computing precisions.
        - effective_order (bool): If `True`, stop including n-gram orders for which precision is 0.
          This should be `True`, if sentence-level BLEU will be computed. (default: True)
        - trg_lang (str): An optional language code to raise potential tokenizer warnings.
        - num_workers (int): Number of workers for multiprocessing.
        """

        lowercase: bool = False
        force: bool = False
        tokenize: Optional[str] = None
        smooth_method: str = "exp"
        smooth_value: Optional[float] = None
        max_ngram_order: int = 4
        effective_order: bool = True
        trg_lang: str = ""
        num_workers: int = 8

    cfg: Config

    @dataclass
    class AggregatedReference:
        """Aggregated reference representation.

        - ngrams (Counter[tuple[str, ...]]): Bags of expected n-gram counts.
        - length (float): Expected length of references.
        """

        ngrams: Counter[tuple[str, ...]]
        length: float

    def __init__(self, cfg: MetricBLEU.Config):
        super().__init__(cfg)
        self.scorer = self._initialize_bleu(cfg)

    @staticmethod
    def _initialize_bleu(cfg: MetricBLEU.Config) -> BLEU:
        scorer = BLEU(
            lowercase=cfg.lowercase,
            force=cfg.force,
            tokenize=cfg.tokenize,
            smooth_method=cfg.smooth_method,
            smooth_value=cfg.smooth_value,
            max_ngram_order=cfg.max_ngram_order,
            effective_order=cfg.effective_order,
            trg_lang=cfg.trg_lang,
        )
        MetricBLEU._score_worker.scorer = scorer
        return scorer

    def score(self, hypothesis: str, reference: str, *_, **__) -> float:
        """Calculate the score of the given hypothesis.

        Args:
            hypothesis (str): Hypothesis.
            reference (str): Reference.

        Returns:
            float: The score of the given hypothesis.
        """
        return self.scorer.sentence_score(hypothesis, [reference]).score

    @staticmethod
    def _score_worker(hypothesis: str, reference: str, *_, **__) -> float:
        """Calculate the score of the given hypothesis.

        Beacause ja-mecab tokenizer cannot be pickled, this method is necessary to use
        multiprocessing.

        Args:
            hypothesis (str): Hypothesis.
            reference (str): Reference.

        Returns:
            float: The score of the given hypothesis.

        Todo:
            - Replace this method with a better logic.
        """
        return MetricBLEU._score_worker.scorer.sentence_score(
            hypothesis, [reference]
        ).score

    def scores(self, hypotheses: list[str], references: list[str], *_, **__) -> Tensor:
        """Calculate the scores of the given hypotheses.

        Args:
            hypotheses (list[str]): N hypotheses.
            references (list[str]): N references.

        Returns:
            Tensor: The N scores of the given hypotheses.
        """
        with concurrent.futures.ProcessPoolExecutor(
            max_workers=self.cfg.num_workers,
            initializer=self._initialize_bleu,
            initargs=(self.cfg,),
        ) as executor:
            with timer.measure("score") as t:
                t.set_delta_ncalls(len(hypotheses))
                return Tensor(
                    list(
                        executor.map(
                            self._score_worker,
                            hypotheses,
                            references,
                            chunksize=math.ceil(len(hypotheses) / self.cfg.num_workers),
                        )
                    )
                )

    def pairwise_scores(
        self, hypotheses: list[str], references: list[str], *_, **__
    ) -> Tensor:
        """Calculate the pairwise scores.

        Args:
            hypotheses (list[str]): Hypotheses.
            references (list[str]): References.

        Returns:
            Tensor: Score matrix of shape `(H, R)`, where `H` is the number
              of hypotheses and `R` is the number of references.
        """
        with concurrent.futures.ProcessPoolExecutor(
            max_workers=self.cfg.num_workers,
            initializer=self._initialize_bleu,
            initargs=(self.cfg,),
        ) as executor:
            with timer.measure("score") as t:
                t.set_delta_ncalls(len(hypotheses) * len(references))

                return Tensor(
                    list(
                        executor.map(
                            self._score_worker,
                            *zip(*itertools.product(hypotheses, references)),
                            chunksize=len(hypotheses),
                        )
                    )
                ).view(len(hypotheses), len(references))

    def corpus_score(
        self, hypotheses: list[str], references: list[str], *_, **__
    ) -> float:
        """Calculate the corpus-level score.

        Args:
            hypotheses (list[str]): Hypotheses.
            references (list[str]): References.

        Returns:
            float: The corpus score.
        """
        return self.scorer.corpus_score(hypotheses, [references]).score

    @staticmethod
    def _compute_bleu(
        correct: list[float],
        total: list[float],
        sys_len: float,
        ref_len: float,
        smooth_method: str = "none",
        smooth_value: Optional[float] = None,
        effective_order: bool = False,
        max_ngram_order: int = MAX_NGRAM_ORDER,
    ) -> float:
        """Computes BLEU score from its sufficient statistics with smoothing.

        Smoothing methods (citing "A Systematic Comparison of Smoothing Techniques for Sentence-Level BLEU",
        Boxing Chen and Colin Cherry, WMT 2014: http://aclweb.org/anthology/W14-3346)

        - none: No smoothing.
        - floor: Method 1 (requires small positive value (0.1 in the paper) to be set)
        - add-k: Method 2 (Generalizing Lin and Och, 2004)
        - exp: Method 3 (NIST smoothing method i.e. in use with mteval-v13a.pl)

        This method extends the original sacrebleu implementation to treat expected n-grams.

        Args:
            correct (list[float]): List of counts of correct ngrams, 1 <= n <= max_ngram_order.
            total (list[float]): List of counts of total ngrams, 1 <= n <= max_ngram_order
            sys_len (float): The cumulative system length
            ref_len (float): The cumulative reference length
            smooth_method (str): The smoothing method to use ('floor', 'add-k', 'exp' or 'none')
            smooth_value (float, optional): The smoothing value for `floor` and `add-k` methods. `None` falls back to default value.
            effective_order (bool): If `True`, stop including n-gram orders for which precision is 0. This should be
                `True`, if sentence-level BLEU will be computed.
            max_ngram_order (int): If given, it overrides the maximum n-gram order (default: 4) when computing precisions.

        Returns:
            float: A BLEU score.
        """
        assert (
            smooth_method in BLEU.SMOOTH_DEFAULTS.keys()
        ), "Unknown smooth_method {smooth_method!r}"

        # Fetch the default value for floor and add-k
        if smooth_value is None:
            smooth_value = BLEU.SMOOTH_DEFAULTS[smooth_method]

        # Compute brevity penalty
        if sys_len < ref_len:
            bp = math.exp(1 - ref_len / sys_len) if sys_len > 0 else 0.0
        else:
            bp = 1.0

        # n-gram precisions
        precisions = [0.0 for x in range(max_ngram_order)]

        # Early stop if there are no matches (#141)
        if not any(correct):
            return 0.0

        smooth_mteval = 1.0
        eff_order = max_ngram_order
        for n in range(1, len(precisions) + 1):
            if smooth_method == "add-k" and n > 1:
                correct[n - 1] += smooth_value
                total[n - 1] += smooth_value

            if total[n - 1] == 0:
                break

            # If the system guesses no i-grams, 1 <= i <= max_ngram_order,
            # the BLEU score is 0 (technically undefined). This is a problem for sentence
            # level BLEU or a corpus of short sentences, where systems will get
            # no credit if sentence lengths fall under the max_ngram_order threshold.
            # This fix scales max_ngram_order to the observed maximum order.
            # It is only available through the API and off by default
            if effective_order:
                eff_order = n

            if correct[n - 1] == 0:
                if smooth_method == "exp":
                    smooth_mteval *= 2
                    precisions[n - 1] = 100.0 / (smooth_mteval * total[n - 1])
                elif smooth_method == "floor":
                    precisions[n - 1] = 100.0 * smooth_value / total[n - 1]
            else:
                precisions[n - 1] = 100.0 * correct[n - 1] / total[n - 1]

        # Compute BLEU score
        score = bp * math.exp(
            sum(
                [
                    math.log(p) if p > 0.0 else -9999999999.0
                    for p in precisions[:eff_order]
                ]
            )
            / eff_order
        )

        return score

    def _aggregate_references(
        self, references: list[str], reference_lprobs: Optional[Tensor] = None
    ) -> AggregatedReference:
        """Aggregate references.

        Args:
            references (list[str]): References.
            reference_lprobs (Tensor, optional): Log-probabilities for each reference sample.
              The shape must be `(len(references),)`. See `https://arxiv.org/abs/2311.05263`.

        Returns:
            MetricBLEU.AggregatedReference: Aggregated reference representation.
        """
        num_references = len(references)
        if reference_lprobs is not None:
            lprobs = reference_lprobs.log_softmax(dim=-1, dtype=torch.float32).tolist()
        else:
            lprobs = [-math.log(num_references)] * num_references

        reference_stats = self.scorer._cache_references([references])
        reference_ngrams: list[Counter[tuple[str, ...]]] = [
            stat["ref_ngrams"] for stat in reference_stats
        ]

        expected_reference_length = sum(
            [
                math.exp(math.log(stat["ref_lens"][0]) + lprob)
                if stat["ref_lens"][0] > 0.0
                else 0.0
                for stat, lprob in zip(reference_stats, lprobs)
            ]
        )

        acc_ngrams: Counter[tuple[str, ...]] = Counter()
        for i, ngrams in enumerate(reference_ngrams):
            for ngram in ngrams:
                # Note: Counter has float values.
                ngrams[ngram] = math.exp(math.log(ngrams[ngram]) + lprobs[i])
            acc_ngrams += ngrams

        return self.AggregatedReference(acc_ngrams, expected_reference_length)

    def expected_scores_reference_aggregation(
        self,
        hypotheses: list[str],
        references: list[str],
        source: Optional[str] = None,
        reference_lprobs: Optional[Tensor] = None,
    ) -> Tensor:
        """Calculate the expected scores for each hypothesis.

        Args:
            hypotheses (list[str]): Hypotheses.
            references (list[str]): References.
            source (str, optional): A source.
            reference_lprobs (Tensor, optional): Log-probabilities for each reference sample.
              The shape must be `(len(references),)`. See `https://arxiv.org/abs/2311.05263`.

        Returns:
            Tensor: The expected scores for each hypothesis.
        """
        with timer.measure("aggregate/references"):
            aggregated_reference = self._aggregate_references(
                references, reference_lprobs=reference_lprobs
            )

        expected_scores = torch.zeros((len(hypotheses),))
        for i, hypothesis in enumerate(hypotheses):
            with timer.measure("expectation"):
                hypothesis = self.scorer._preprocess_segment(hypothesis)
                # Extract n-grams for the hypothesis
                hyp_ngrams, hyp_len = extract_all_word_ngrams(
                    hypothesis, 1, self.scorer.max_ngram_order
                )

                # Count the stats
                # Although counter has its internal & and | operators, this is faster
                correct = [0.0 for i in range(self.scorer.max_ngram_order)]
                total = correct[:]
                for hyp_ngram, hyp_count in hyp_ngrams.items():
                    # n-gram order
                    n = len(hyp_ngram) - 1
                    # count hypothesis n-grams
                    total[n] += float(hyp_count)
                    # count matched n-grams
                    if hyp_ngram in aggregated_reference.ngrams:
                        correct[n] += float(
                            min(hyp_count, aggregated_reference.ngrams[hyp_ngram])
                        )

                stats = [hyp_len, aggregated_reference.length] + correct + total
                expected_scores[i] = self._compute_bleu(
                    correct=stats[2 : 2 + self.scorer.max_ngram_order],
                    total=stats[2 + self.scorer.max_ngram_order :],
                    sys_len=float(stats[0]),
                    ref_len=float(stats[1]),
                    smooth_method=self.scorer.smooth_method,
                    smooth_value=self.scorer.smooth_value,
                    effective_order=self.scorer.effective_order,
                    max_ngram_order=self.scorer.max_ngram_order,
                )

        return expected_scores



from typing import Optional

import pytest
import torch
from torch import Tensor

from .functional import expectation

H = 8
R = 5

A = torch.rand(H, R)


@pytest.mark.parametrize(
    "lprobs", [None, torch.FloatTensor([-3.0, -2.4, -1.9, -5.5, -4.3])]
)
def test_expectation(lprobs: Optional[Tensor]):
    e = expectation(A, lprobs=lprobs)
    assert list(e.shape) == [H]
    if lprobs is None:
        assert torch.allclose(e, A.mean(dim=-1))
    else:
        with pytest.raises(ValueError):
            expectation(torch.rand(H, R - 1), lprobs=lprobs)

        weights = lprobs.exp() / lprobs.exp().sum(dim=-1, keepdim=True)
        assert torch.allclose(e, (A @ weights[None, :].T).squeeze(-1))



import pytest
import torch

from mbrs.decoders.aggregate_mbr import DecoderAggregateMBR
from mbrs.metrics.comet import MetricCOMET
from mbrs.selectors.base import Selector

from .centroid_mbr import DecoderCentroidMBR

SOURCE = [
    "これはテストです",
    "これはテストです",
    "これはテストです",
    "これはテストです",
]
HYPOTHESES = [
    ["another test", "this is a test", "this is a fest"],
    ["another test", "this is a fest", "this is a test"],
    ["this is a test"],
    ["Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;"],
]
REFERENCES = [
    ["another test", "this is a test", "this is a fest"],
    ["this is a test", "ref", "these are tests", "this is the test"],
    ["this is a test"],
    ["producţia de zahăr brut se exprimă în zahăr alb;"],
]

BEST_INDICES = [1, 2, 0, 0]
BEST_SENTENCES = [
    "this is a test",
    "this is a test",
    "this is a test",
    "Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;",
]
SCORES = torch.Tensor([0.88974, 0.76127, 0.99257, 0.78060])

NCENTROIDS = 2


class TestDecoderCBMBR:
    @pytest.mark.parametrize("kmeanspp", [True, False])
    def test_decode(self, metric_comet: MetricCOMET, kmeanspp: bool):
        decoder = DecoderCentroidMBR(
            DecoderCentroidMBR.Config(ncentroids=NCENTROIDS, kmeanspp=kmeanspp),
            metric_comet,
        )
        for i, (hyps, refs) in enumerate(zip(HYPOTHESES, REFERENCES)):
            output = decoder.decode(hyps, refs, SOURCE[i], nbest=1)
            assert output.idx[0] == BEST_INDICES[i]
            assert output.sentence[0] == BEST_SENTENCES[i]
            assert torch.isclose(
                torch.tensor(output.score[0]),
                SCORES[i],
                atol=0.0005 / 100,
            )

            output = decoder.decode(
                hyps,
                refs,
                SOURCE[i],
                nbest=1,
                reference_lprobs=torch.Tensor([-2.000]).repeat(len(refs)),
            )
            assert output.idx[0] == BEST_INDICES[i]
            assert output.sentence[0] == BEST_SENTENCES[i]
            assert torch.isclose(
                torch.tensor(output.score[0]),
                SCORES[i],
                atol=0.0005 / 100,
            )

    @pytest.mark.parametrize("nbest", [1, 2])
    @pytest.mark.parametrize("kmeanspp", [True, False])
    def test_decode_equivalent_with_aggregate(
        self, metric_comet: MetricCOMET, nbest: int, kmeanspp: bool, selector: Selector
    ):
        decoder_cbmbr = DecoderCentroidMBR(
            DecoderCentroidMBR.Config(ncentroids=1, kmeanspp=kmeanspp),
            metric_comet,
            selector=selector,
        )
        decoder_aggregate = DecoderAggregateMBR(
            DecoderAggregateMBR.Config(), metric_comet, selector=selector
        )
        for i, (hyps, refs) in enumerate(zip(HYPOTHESES, REFERENCES)):
            output_cbmbr = decoder_cbmbr.decode(hyps, refs, SOURCE[i], nbest=nbest)
            output_aggregate = decoder_aggregate.decode(
                hyps, refs, SOURCE[i], nbest=nbest
            )
            assert output_cbmbr.idx == output_aggregate.idx
            assert output_cbmbr.sentence == output_aggregate.sentence
            torch.testing.assert_close(
                torch.tensor(output_cbmbr.score),
                torch.tensor(output_aggregate.score),
                rtol=1e-6,
                atol=0.0005 / 100,
            )

    @pytest.mark.parametrize("nbest", [1, 2])
    def test_decode_selector(
        self, metric_comet: MetricCOMET, nbest: int, selector: Selector
    ):
        decoder = DecoderCentroidMBR(
            DecoderCentroidMBR.Config(ncentroids=NCENTROIDS),
            metric_comet,
            selector=selector,
        )
        for i, (hyps, refs) in enumerate(zip(HYPOTHESES, REFERENCES)):
            output = decoder.decode(hyps, refs, SOURCE[i], nbest=nbest)
            assert len(output.sentence) == min(nbest, len(hyps))
            assert len(output.score) == min(nbest, len(hyps))

            output = decoder.decode(
                hyps,
                refs,
                SOURCE[i],
                nbest=nbest,
                reference_lprobs=torch.Tensor([-2.000]).repeat(len(refs)),
            )
            assert len(output.sentence) == min(nbest, len(hyps))
            assert len(output.score) == min(nbest, len(hyps))



import argparse
import logging
import sys
from argparse import Namespace
from pathlib import Path
from typing import Sequence

import simple_parsing
from simple_parsing.wrappers.field_wrapper import ArgumentGenerationMode, FieldWrapper

logger = logging.getLogger(__name__)


class ArgumentParser(simple_parsing.ArgumentParser):
    def preprocess_parser(self) -> None:
        """Preprocess ArgumentParser."""
        self.parse_known_args_preprocess(sys.argv[1:])

    def parse_known_args_preprocess(
        self,
        args: Sequence[str] | None = None,
        namespace: Namespace | None = None,
        attempt_to_reorder: bool = False,
    ) -> None:
        # default Namespace built from parser defaults
        if namespace is None:
            namespace = Namespace()
        if self.config_path:
            if isinstance(self.config_path, Path):
                config_paths = [self.config_path]
            else:
                config_paths = self.config_path
            for config_file in config_paths:
                self.set_defaults(config_file)

        if self.add_config_path_arg:
            temp_parser = ArgumentParser(
                add_config_path_arg=False,
                add_help=False,
                add_option_string_dash_variants=FieldWrapper.add_dash_variants,
                argument_generation_mode=FieldWrapper.argument_generation_mode,
                nested_mode=FieldWrapper.nested_mode,
            )
            temp_parser.add_argument(
                "--config_path",
                type=Path,
                default=self.config_path,
                help="Path to a config file containing default values to use.",
            )
            args_with_config_path, args = temp_parser.parse_known_args(args)
            config_path = args_with_config_path.config_path

            if config_path is not None:
                config_paths = (
                    config_path if isinstance(config_path, list) else [config_path]
                )
                for config_file in config_paths:
                    self.set_defaults(config_file)

            # Adding it here just so it shows up in the help message. The default will be set in
            # the help string.
            if self._option_string_actions.get("--config_path", None) is None:
                self.add_argument(
                    "--config_path",
                    type=Path,
                    default=config_path,
                    help="Path to a config file containing default values to use.",
                )

        assert isinstance(args, list)
        self._preprocessing(args=args, namespace=namespace)

    def parse_known_args(
        self,
        args: Sequence[str] | None = None,
        namespace: Namespace | None = None,
        attempt_to_reorder: bool = False,
    ):
        # NOTE: since the usual ArgumentParser.parse_args() calls
        # parse_known_args, we therefore just need to overload the
        # parse_known_args method to support both.
        if args is None:
            # args default to the system args
            args = sys.argv[1:]
        else:
            # make sure that args are mutable
            args = list(args)

        self.parse_known_args_preprocess(
            args=args, namespace=namespace, attempt_to_reorder=attempt_to_reorder
        )
        logger.debug(
            f"Parser {id(self)} is parsing args: {args}, namespace: {namespace}"
        )
        parsed_args, unparsed_args = super(
            simple_parsing.ArgumentParser, self
        ).parse_known_args(args, namespace)

        if unparsed_args and self._subparsers and attempt_to_reorder:
            logger.warning(
                f"Unparsed arguments when using subparsers. Will "
                f"attempt to automatically re-order the unparsed arguments "
                f"{unparsed_args}."
            )
            index_in_start = args.index(unparsed_args[0])
            # Simply 'cycle' the args to the right ordering.
            new_start_args = args[index_in_start:] + args[:index_in_start]
            parsed_args, unparsed_args = super(
                simple_parsing.ArgumentParser, self
            ).parse_known_args(new_start_args)

        parsed_args = self._postprocessing(parsed_args)
        return parsed_args, unparsed_args


class DataclassWrapper(simple_parsing.wrappers.DataclassWrapper):
    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        if self._name == "common":
            FieldWrapper.argument_generation_mode = ArgumentGenerationMode.FLAT
        else:
            FieldWrapper.argument_generation_mode = ArgumentGenerationMode.NESTED
        super().add_arguments(parser)
        FieldWrapper.argument_generation_mode = ArgumentGenerationMode.NESTED



import pathlib

import yaml

from mbrs.cli.decode import get_argparser


class TestArgumentParser:
    def test_config_load(self, tmp_path: pathlib.Path):
        config_path = tmp_path / "config.yaml"
        hyps_path = tmp_path / "hyps.txt"
        cfg_dict = {
            "common": {
                "metric": "chrf",
                "decoder": "probabilistic_mbr",
                "num_candidates": 2,
            },
            "metric": {"char_order": 4},
            "decoder": {"reduction_factor": 16.0},
        }

        with open(config_path, mode="w") as f:
            yaml.dump(cfg_dict, f)

        with open(hyps_path, mode="w") as f:
            f.writelines(["tests", "a test"])

        cmd_args = [str(hyps_path), "--config_path", str(config_path)]
        parser = get_argparser(cmd_args)
        args = parser.parse_args(args=cmd_args)
        for key, value in cfg_dict["common"].items():
            assert getattr(args.common, key) == value
        for key, value in cfg_dict["metric"].items():
            assert getattr(args.metric, key) == value
        for key, value in cfg_dict["decoder"].items():
            assert getattr(args.decoder, key) == value



from __future__ import annotations

import enum

from mbrs import registry

from .base import (
    Metric,
    MetricAggregatable,
    MetricBase,
    MetricCacheable,
    MetricReferenceless,
)

register, get_metric = registry.setup("metric")

from .bleu import MetricBLEU
from .bleurt import MetricBLEURT
from .chrf import MetricChrF
from .comet import MetricCOMET
from .cometkiwi import MetricCOMETkiwi
from .ter import MetricTER
from .xcomet import MetricXCOMET

__all__ = [
    "MetricBase",
    "Metric",
    "MetricAggregatable",
    "MetricCacheable",
    "MetricReferenceless",
    "MetricBLEU",
    "MetricChrF",
    "MetricCOMET",
    "MetricCOMETkiwi",
    "MetricTER",
    "MetricXCOMET",
    "MetricBLEURT",
]


class MetricEnum(str, enum.Enum): ...


Metrics = MetricEnum("Metrics", {k: k for k in registry.get_registry("metric").keys()})



import pytest
import torch

from .bleurt import MetricBLEURT

HYPOTHESES = [
    "this is a test",
    "another test",
    "this is a fest",
    "Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;",
]
REFERENCES = [
    "ref",
    "this is a test",
    "producţia de zahăr brut se exprimă în zahăr alb;",
]
SCORES = torch.Tensor(
    [
        [0.10943, 0.97992, 0.08830],
        [0.06851, 0.59438, 0.03788],
        [0.12863, 0.43761, 0.08151],
        [0.16558, 0.17050, 0.65812],
    ]
)


@pytest.fixture(scope="module")
def metric_bleurt():
    return MetricBLEURT(MetricBLEURT.Config())


class TestMetricBLEURT:
    def test_score(self, metric_bleurt: MetricBLEURT):
        for i, hyp in enumerate(HYPOTHESES):
            for j, ref in enumerate(REFERENCES):
                assert torch.isclose(
                    SCORES[i, j],
                    torch.tensor(metric_bleurt.score(hyp, ref)),
                    atol=0.0005 / 100,
                )

    def test_scores(self, metric_bleurt: MetricBLEURT):
        hyps = ["another test", "this is a test", "this is an test"]
        refs = ["another test", "this is a fest", "this is a test"]
        torch.testing.assert_close(
            metric_bleurt.scores(hyps, refs).cpu().float(),
            torch.FloatTensor([0.98427, 0.46749, 0.88140]),
            atol=0.0005 / 100,
            rtol=1e-6,
        )

    def test_pairwise_scores(self, metric_bleurt: MetricBLEURT):
        pairwise_scores = metric_bleurt.pairwise_scores(HYPOTHESES, REFERENCES)
        torch.testing.assert_close(
            pairwise_scores,
            SCORES.to(metric_bleurt.device),
            atol=0.0005 / 100,
            rtol=1e-6,
        )

    def test_pairwise_scores_empty_inputs(self, metric_bleurt: MetricBLEURT):
        pairwise_scores = metric_bleurt.pairwise_scores(
            ["this is a test", ""], ["", "this is a test", ""]
        )
        assert tuple(pairwise_scores.shape) == (2, 3)
        torch.testing.assert_close(
            pairwise_scores[0][1],
            torch.tensor(0.97992).to(metric_bleurt.device),
            atol=0.0005 / 100,
            rtol=1e-6,
        )

    def test_expected_scores(self, metric_bleurt: MetricBLEURT):
        expected_scores = metric_bleurt.expected_scores(HYPOTHESES, REFERENCES)
        torch.testing.assert_close(
            expected_scores,
            SCORES.mean(dim=1).to(metric_bleurt.device),
            atol=0.0005 / 100,
            rtol=1e-6,
        )

    def test_corpus_score(self, metric_bleurt: MetricBLEURT):
        hyps = ["another test", "this is a test", "this is an test"]
        refs = ["another test", "this is a fest", "this is a test"]
        assert torch.isclose(
            torch.tensor(metric_bleurt.corpus_score(hyps, refs)),
            torch.FloatTensor([0.98427, 0.46749, 0.88140]).mean(),
        )



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional

import torch
from torch import Tensor

from mbrs import functional, timer
from mbrs.metrics import Metric, MetricCacheable
from mbrs.selectors import Selector, SelectorNbest

from . import register
from .mbr import DecoderMBR


@register("pruning_mbr")
class DecoderPruningMBR(DecoderMBR):
    """Pruning MBR decoder class.

    References:
        J. Cheng and A. Vlachos, 2023,
        "Faster Minimum Bayes Risk Decoding with Confidence-based Pruning".
        https://aclanthology.org/2023.emnlp-main.767/
    """

    def __init__(
        self,
        cfg: DecoderPruningMBR.Config,
        metric: Metric,
        selector: Selector = SelectorNbest(SelectorNbest.Config()),
    ) -> None:
        if not isinstance(selector, SelectorNbest):
            raise ValueError(
                "Confidence-based pruning cannot be combined with other selectors than the nbest."
            )
        super().__init__(cfg, metric, selector)

    @dataclass
    class Config(DecoderMBR.Config):
        """Configuration for the decoder.

        - alpha (float): Prune hypotheses based on this confidence threshold.
        - sampling_shceduler (list[int]): Sample size scheduler. For each step, the
          number of samples will be the t-th number.
        - num_boostrap_samples (int): Number of boostrap samples.
        - seed (int): Random seed for bootstrap sampling.
        """

        alpha: float = 0.99
        sampling_scheduler: list[int] = field(
            default_factory=lambda: [8, 16, 32, 64, 128, 256]
        )
        num_bootstrap_samples: int = 500
        seed: int = 0

    cfg: Config

    def decode_pruning(
        self,
        hypotheses: list[str],
        references: list[str],
        source: Optional[str] = None,
        nbest: int = 1,
        reference_lprobs: Optional[Tensor] = None,
    ) -> tuple[list[float], list[int]]:
        """Select the n-best hypotheses using pruning MBR decoding.

        Args:
            hypotheses (list[str]): Hypotheses.
            references (list[str]): References.
            source (str, optional): A source.
            nbest (int): Return the n-best hypotheses.
            reference_lprobs (Tensor, optional): Log-probabilities for each reference sample.
              The shape must be `(len(references),)`. See `https://arxiv.org/abs/2311.05263`.

        Returns:
            - list[float]: Top-k scores.
            - list[int]: Top-k indices.
        """
        rng = torch.Generator(device=self.metric.device).manual_seed(self.cfg.seed)
        H = len(hypotheses)
        max_r = min(max(self.cfg.sampling_scheduler), len(references))
        pairwise_scores = torch.zeros((H, max_r), device=self.metric.device)
        orig_indices = torch.arange(H, device=self.metric.device)

        if isinstance(self.metric, MetricCacheable):
            with timer.measure("encode/hypotheses"):
                hypotheses_ir = self.metric.encode(hypotheses)
            references_ir = hypotheses_ir if hypotheses == references else None
            if source is None:
                source_ir = None
            else:
                with timer.measure("encode/source"):
                    source_ir = self.metric.encode([source])

        with timer.measure("pruning_mbr"):
            # Algorithm 1 in the paper.
            prev_r = 0
            for t, r in enumerate(self.cfg.sampling_scheduler):
                r = min(r, len(references))
                if r <= prev_r:
                    break

                # Equation 5 and Algorithm 2 in the paper.
                if isinstance(self.metric, MetricCacheable):
                    if references_ir is None:
                        with timer.measure("encode/references"):
                            references_ir_t = self.metric.encode(references[prev_r:r])
                    else:
                        references_ir_t = references_ir[prev_r:r]

                    pairwise_scores[:, prev_r:r] = self.metric.pairwise_scores_from_ir(
                        hypotheses_ir, references_ir_t, source_ir
                    )
                else:
                    pairwise_scores[:, prev_r:r] = self.metric.pairwise_scores(
                        hypotheses, references[prev_r:r], source
                    )

                expected_scores = functional.expectation(
                    pairwise_scores[:, :r],
                    lprobs=reference_lprobs[:r]
                    if reference_lprobs is not None
                    else None,
                )
                current_best_idx = self.argbest(expected_scores)
                sample_indices = torch.randint(
                    r,
                    size=(self.cfg.num_bootstrap_samples, r),
                    device=self.metric.device,
                    generator=rng,
                )
                bootstrap_expected_scores = functional.expectation(
                    pairwise_scores[:, sample_indices],
                    lprobs=reference_lprobs[sample_indices]
                    if reference_lprobs is not None
                    else None,
                )
                num_wins = (
                    (
                        bootstrap_expected_scores
                        >= bootstrap_expected_scores[current_best_idx]
                    )
                    if self.maximize
                    else (
                        bootstrap_expected_scores
                        <= bootstrap_expected_scores[current_best_idx]
                    )
                )
                win_rates = num_wins.float().mean(dim=1)
                winners = (win_rates > 1 - self.cfg.alpha).nonzero(as_tuple=True)[0]
                num_winners = len(winners)
                if num_winners >= nbest:
                    if isinstance(self.metric, MetricCacheable):
                        hypotheses_ir = hypotheses_ir.index_select(0, winners)
                    else:
                        hypotheses = [hypotheses[i] for i in winners]
                    pairwise_scores = pairwise_scores[winners]
                    orig_indices = orig_indices[winners]
                    prev_r = r
                else:
                    break
            expected_scores = functional.expectation(
                pairwise_scores[:, :prev_r],
                lprobs=reference_lprobs[:prev_r]
                if reference_lprobs is not None
                else None,
            )

        topk_scores, topk_indices = self.topk(expected_scores, k=nbest)
        return topk_scores, orig_indices[topk_indices].tolist()

    def decode(
        self,
        hypotheses: list[str],
        references: list[str],
        source: Optional[str] = None,
        nbest: int = 1,
        reference_lprobs: Optional[Tensor] = None,
    ) -> DecoderMBR.Output:
        """Select the n-best hypotheses based on the strategy.

        Args:
            hypotheses (list[str]): Hypotheses.
            references (list[str]): References.
            source (str, optional): A source.
            nbest (int): Return the n-best hypotheses.
            reference_lprobs (Tensor, optional): Log-probabilities for each reference sample.
              The shape must be `(len(references),)`. See `https://arxiv.org/abs/2311.05263`.

        Returns:
            DecoderMBR.Output: The n-best hypotheses.
        """

        topk_scores, topk_indices = self.decode_pruning(
            hypotheses,
            references,
            source,
            nbest=nbest,
            reference_lprobs=reference_lprobs,
        )
        return self.Output(
            idx=topk_indices,
            sentence=[hypotheses[idx] for idx in topk_indices],
            score=topk_scores,
        )



from __future__ import annotations

import math
from dataclasses import dataclass
from typing import Optional

import torch
from torch import Tensor

from mbrs import functional, timer
from mbrs.metrics import MetricCacheable
from mbrs.modules.als import MatrixFactorizationALS

from . import register
from .mbr import DecoderMBR


@register("probabilistic_mbr")
class DecoderProbabilisticMBR(DecoderMBR):
    """Probabilistic MBR decoder using alternating least squares (ALS) approximation.

    References:
        F. Trabelsi et al., 2024,
        "Efficient Minimum Bayes Risk Decoding using Low-Rank Matrix Completion Algorithms".
        https://arxiv.org/abs/2406.02832
    """

    cfg: Config

    @dataclass
    class Config(DecoderMBR.Config):
        """Configuration for the decoder.

        - reduction_factor (float): Reduction factor.
          The computational budget will be reduced to `1 / reduction_factor`.
        - regularization_weight (float): Weight of L2 regularization.
        - rank (int): Rank of the factarized matrices.
        - niter (int): The number of alternating steps performed.
        - seed (int): Random seed.
        """

        reduction_factor: float = 8.0
        regularization_weight: float = 0.1
        rank: int = 8
        niter: int = 10
        seed: int = 0

    def pairwise_scores_probabilistic(
        self,
        hypotheses: list[str],
        references: list[str],
        source: Optional[str] = None,
    ) -> Tensor:
        """Compute approximated pairwise scores using the probabilistic MBR algorithm.

        Args:
            hypotheses (list[str]): Hypotheses.
            references (list[str]): References.
            source (str, optional): A source.

        Returns:
            Tensor: Approximated pairwise scores of shape `(H, R)`.
        """
        rng = torch.Generator().manual_seed(self.cfg.seed)
        H = len(hypotheses)
        R = len(references)
        num_ucalcs = math.ceil(H * R / self.cfg.reduction_factor)

        pairwise_scores = torch.zeros((H, R), device=self.metric.device)
        pairwise_sample_indices = torch.randperm(H * R, generator=rng)[:num_ucalcs]
        hypothesis_sample_indices: list[int] = (pairwise_sample_indices // R).tolist()
        reference_sample_indices: list[int] = (pairwise_sample_indices % R).tolist()
        hypothesis_samples = [hypotheses[i] for i in hypothesis_sample_indices]
        reference_samples = [references[j] for j in reference_sample_indices]

        # For COMET-22
        if isinstance(self.metric, MetricCacheable):
            hypothesis_sample_indices_set = set(hypothesis_sample_indices)
            reference_sample_indices_set = set(reference_sample_indices)
            hypothesis_samples_deduped = [
                hypotheses[i] for i in hypothesis_sample_indices_set
            ]
            reference_samples_deduped = [
                references[j] for j in reference_sample_indices_set
            ]
            with timer.measure("encode/hypotheses"):
                ir = self.metric.encode(hypothesis_samples_deduped)
                hypotheses_ir = ir.new_zeros((H, self.metric.embed_dim))
                references_ir = ir.new_zeros((R, self.metric.embed_dim))
                hypotheses_ir[list(hypothesis_sample_indices_set)] = ir
            with timer.measure("encode/references"):
                if hypotheses == references:
                    seen_indices = list(
                        hypothesis_sample_indices_set & reference_sample_indices_set
                    )
                    unseen_indices = list(
                        reference_sample_indices_set - hypothesis_sample_indices_set
                    )
                    if len(seen_indices) > 0:
                        references_ir[seen_indices] = hypotheses_ir[seen_indices]
                    if len(unseen_indices) > 0:
                        references_ir[unseen_indices] = self.metric.encode(
                            [references[j] for j in unseen_indices]
                        )
                else:
                    references_ir[list(reference_sample_indices_set)] = (
                        self.metric.encode(reference_samples_deduped)
                    )
            if source is None:
                source_ir = None
            else:
                with timer.measure("encode/source"):
                    source_ir = self.metric.encode([source])

        # Algorithm 2 in the paper.
        if isinstance(self.metric, MetricCacheable):
            num_hyp_samples = len(hypothesis_sample_indices)
            for i in range(0, num_hyp_samples, H):
                pairwise_scores[
                    hypothesis_sample_indices[i : i + H],
                    reference_sample_indices[i : i + H],
                ] = self.metric.scores_from_ir(
                    hypotheses_ir[hypothesis_sample_indices[i : i + H]],
                    references_ir[reference_sample_indices[i : i + H]],
                    source_ir.repeat(min(H, num_hyp_samples - i), 1)
                    if source_ir is not None
                    else None,
                ).float()
        else:
            pairwise_scores[hypothesis_sample_indices, reference_sample_indices] = (
                self.metric.scores(
                    hypothesis_samples,
                    reference_samples,
                    [source] * len(hypothesis_samples) if source is not None else None,
                ).float()
            )
        observed_mask = pairwise_scores.new_zeros((H, R), dtype=torch.bool)
        observed_mask[hypothesis_sample_indices, reference_sample_indices] = True

        # Algorithm 1 in the paper.
        als = MatrixFactorizationALS(
            regularization_weight=self.cfg.regularization_weight, rank=self.cfg.rank
        )
        X, Y = als.factorize(
            pairwise_scores,
            observed_mask=observed_mask,
            niter=self.cfg.niter,
            seed=self.cfg.seed,
        )
        reconstructed_pairwise_scores = X @ Y.T
        return reconstructed_pairwise_scores

    def decode(
        self,
        hypotheses: list[str],
        references: list[str],
        source: Optional[str] = None,
        nbest: int = 1,
        reference_lprobs: Optional[Tensor] = None,
    ) -> DecoderMBR.Output:
        """Select the n-best hypotheses based on the strategy.

        Args:
            hypotheses (list[str]): Hypotheses.
            references (list[str]): References.
            source (str, optional): A source.
            nbest (int): Return the n-best hypotheses.
            reference_lprobs (Tensor, optional): Log-probabilities for each reference sample.
              The shape must be `(len(references),)`. See `https://arxiv.org/abs/2311.05263`.

        Returns:
            DecoderMBR.Output: The n-best hypotheses.
        """

        if self.cfg.reduction_factor <= 1.0:
            expected_scores = self.metric.expected_scores(
                hypotheses, references, source, reference_lprobs=reference_lprobs
            )
        else:  # Probabilistic MBR decoding
            pairwise_scores = self.pairwise_scores_probabilistic(
                hypotheses, references, source
            )
            expected_scores = functional.expectation(
                pairwise_scores, lprobs=reference_lprobs
            )

        selector_outputs = self.select(
            hypotheses, expected_scores, nbest=nbest, source=source
        )
        return (
            self.Output(
                idx=selector_outputs.idx,
                sentence=selector_outputs.sentence,
                score=selector_outputs.score,
            )
            | selector_outputs
        )



from __future__ import annotations

from mbrs import timer

from . import DecoderReferenceless, register


@register("rerank")
class DecoderRerank(DecoderReferenceless):
    """Reranking decoder class.

    - Time complexity: O(N)
    - Space complexity: O(N)
    """

    def decode(
        self, hypotheses: list[str], source: str, nbest: int = 1
    ) -> DecoderRerank.Output:
        """Select the n-best hypotheses based on the strategy.

        Args:
            hypotheses (list[str]): Hypotheses.
            source (str): A source.
            nbest (int): Return the n-best hypotheses.

        Returns:
            DecoderRerank.Output: The n-best hypotheses.
        """
        with timer.measure("rerank"):
            scores = self.metric.scores(hypotheses, sources=[source] * len(hypotheses))

        selector_outputs = self.select(hypotheses, scores, nbest=nbest, source=source)
        return (
            self.Output(
                idx=selector_outputs.idx,
                sentence=selector_outputs.sentence,
                score=selector_outputs.score,
            )
            | selector_outputs
        )



from mbrs import registry

from .base import Selector

register, get_selector = registry.setup("selector")

from .diverse import SelectorDiverse
from .nbest import SelectorNbest

__all__ = ["Selector", "SelectorNbest", "SelectorDiverse"]



import numpy as np
import pytest
import torch

from mbrs.metrics import MetricChrF, MetricCOMET, get_metric
from mbrs.selectors import Selector, SelectorDiverse

from .pruning_mbr import DecoderPruningMBR

SOURCE = [
    "これはテストです",
    "これはテストです",
    "これはテストです",
    "これはテストです",
]
HYPOTHESES = [
    ["another test", "this is a test", "this is a fest", "x", "this is test"],
    ["another test", "this is a fest", "this is a test"],
    ["this is a test"],
    ["Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;"],
]
REFERENCES = [
    ["another test", "this is a test", "this is a fest", "x", "this is test"],
    ["this is a test", "ref", "these are tests", "this is the test"],
    ["this is a test"],
    ["producţia de zahăr brut se exprimă în zahăr alb;"],
]

BEST_INDICES = [1, 2, 0, 0]
BEST_SENTENCES = [
    "this is a test",
    "this is a test",
    "this is a test",
    "Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;",
]
SCORES_COMET = np.array([0.84780, 0.85304, 0.99257, 0.78060], dtype=np.float32)
SCORES_CHRF = np.array([48.912, 44.239, 100.0, 46.161], dtype=np.float32)


class TestDecoderPruningMBR:
    def test_decode_chrf(self):
        metric = MetricChrF(MetricChrF.Config())
        decoder = DecoderPruningMBR(
            DecoderPruningMBR.Config(sampling_scheduler=[2, 4, 8]), metric
        )
        for i, (hyps, refs) in enumerate(zip(HYPOTHESES, REFERENCES)):
            output = decoder.decode(hyps, refs, SOURCE[i], nbest=1)
            assert output.idx[0] == BEST_INDICES[i]
            assert output.sentence[0] == BEST_SENTENCES[i]
            assert np.allclose(
                np.array(output.score[0], dtype=np.float32), SCORES_CHRF[i], atol=0.0005
            )

            output = decoder.decode(
                hyps,
                refs,
                SOURCE[i],
                nbest=1,
                reference_lprobs=torch.Tensor([-2.000]).repeat(len(refs)),
            )
            assert output.idx[0] == BEST_INDICES[i]
            assert output.sentence[0] == BEST_SENTENCES[i]
            assert np.allclose(
                np.array(output.score[0], dtype=np.float32), SCORES_CHRF[i], atol=0.0005
            )

    def test_decode_comet(self, metric_comet: MetricCOMET):
        decoder = DecoderPruningMBR(
            DecoderPruningMBR.Config(sampling_scheduler=[2, 4, 8]), metric_comet
        )
        for i, (hyps, refs) in enumerate(zip(HYPOTHESES, REFERENCES)):
            output = decoder.decode(hyps, refs, SOURCE[i], nbest=1)
            assert output.idx[0] == BEST_INDICES[i]
            assert output.sentence[0] == BEST_SENTENCES[i]
            assert np.allclose(
                np.array(output.score[0], dtype=np.float32),
                SCORES_COMET[i],
                atol=0.0005,
            )

    @pytest.mark.parametrize("metric_type", ["bleu", "ter"])
    @pytest.mark.parametrize("nbest", [1, 2])
    def test_decode_selector(self, metric_type: str, nbest: int, selector: Selector):
        metric_cls = get_metric(metric_type)
        if isinstance(selector, SelectorDiverse):
            with pytest.raises(ValueError):
                DecoderPruningMBR(
                    DecoderPruningMBR.Config(sampling_scheduler=[2, 4, 8]),
                    metric_cls(metric_cls.Config()),
                    selector=selector,
                )
            return

        decoder = DecoderPruningMBR(
            DecoderPruningMBR.Config(sampling_scheduler=[2, 4, 8]),
            metric_cls(metric_cls.Config()),
            selector=selector,
        )
        for i, (hyps, refs) in enumerate(zip(HYPOTHESES, REFERENCES)):
            output = decoder.decode(hyps, refs, nbest=nbest, reference_lprobs=None)
            assert len(output.sentence) == min(nbest, len(hyps))
            assert len(output.score) == min(nbest, len(hyps))
            output = decoder.decode(
                hyps,
                refs,
                nbest=nbest,
                reference_lprobs=torch.Tensor([-2.000]).repeat(len(refs)),
            )
            assert len(output.sentence) == min(nbest, len(hyps))
            assert len(output.score) == min(nbest, len(hyps))



from __future__ import annotations

from typing import Optional

from torch import Tensor

from mbrs.metrics import MetricAggregatable

from . import register
from .mbr import DecoderMBR


@register("aggregate_mbr")
class DecoderAggregateMBR(DecoderMBR):
    """MBR decoding with reference aggregation.

    - Time complexity: O(N)
    - Space complexity: O(N)

    References:
        J. DeNero et al., 2009,
        "Fast Consensus Decoding over Translation Forests".
        https://aclanthology.org/P09-1064/

        J. Vamvas and R. Sennrich, 2024,
        "Linear-time Minimum Bayes Risk Decoding with Reference Aggregation".
        https://arxiv.org/abs/2402.04251
    """

    def decode(
        self,
        hypotheses: list[str],
        references: list[str],
        source: Optional[str] = None,
        nbest: int = 1,
        reference_lprobs: Optional[Tensor] = None,
    ) -> DecoderAggregateMBR.Output:
        """Select the n-best hypotheses based on the strategy.

        Args:
            hypotheses (list[str]): Hypotheses.
            references (list[str]): References.
            source (str, optional): A source.
            nbest (int): Return the n-best hypotheses.
            reference_lprobs (Tensor, optional): Log-probabilities for each reference sample.
              The shape must be `(len(references),)`. See `https://arxiv.org/abs/2311.05263`.

        Returns:
            DecoderAggregateMBR.Output: The n-best hypotheses.
        """
        assert isinstance(self.metric, MetricAggregatable)
        expected_scores = self.metric.expected_scores_reference_aggregation(
            hypotheses, references, source=source, reference_lprobs=reference_lprobs
        )
        selector_outputs = self.select(
            hypotheses, expected_scores, nbest=nbest, source=source
        )
        return (
            self.Output(
                idx=selector_outputs.idx,
                sentence=selector_outputs.sentence,
                score=selector_outputs.score,
            )
            | selector_outputs
        )



from __future__ import annotations

import itertools
from dataclasses import dataclass

import torch
import transformers
from bleurt_pytorch import (
    BleurtForSequenceClassification,
    BleurtTokenizer,
)
from torch import Tensor
from transformers.tokenization_utils import BatchEncoding, EncodedInputPair

from mbrs import timer

from . import Metric, register

transformers.logging.set_verbosity_error()


@register("bleurt")
class MetricBLEURT(Metric):
    """BLEURT metric class.

    We employ the PyTorch port version to implement this metric instead of the original version:
    https://github.com/lucadiliello/bleurt-pytorch
    (thanks to @lucadiliello)

    Available checkpoints:

    - lucadiliello/BLEURT-20
    - lucadiliello/BLEURT-20-D12
    - lucadiliello/BLEURT-20-D3
    - lucadiliello/BLEURT-20-D6
    - lucadiliello/bleurt-base-128
    - lucadiliello/bleurt-base-512
    - lucadiliello/bleurt-large-128
    - lucadiliello/bleurt-large-512
    - lucadiliello/bleurt-tiny-128
    - lucadiliello/bleurt-tiny-512
    """

    scorer: BleurtForSequenceClassification

    @dataclass
    class Config(Metric.Config):
        """BLEURT metric configuration.

        - model (str): Model name or path.
        - batch_size (int): Batch size.
        - fp16 (bool): Use float16 for the forward computation.
        - bf16 (bool): Use bfloat16 for the forward computation.
        - cpu (bool): Use CPU for the forward computation.
        """

        model: str = "lucadiliello/BLEURT-20-D12"
        batch_size: int = 64
        fp16: bool = False
        bf16: bool = False
        cpu: bool = False

    def __init__(self, cfg: MetricBLEURT.Config):
        self.cfg = cfg
        self.scorer = BleurtForSequenceClassification.from_pretrained(cfg.model)
        self.tokenizer = BleurtTokenizer.from_pretrained(cfg.model)
        self.max_length = self.tokenizer.max_model_input_sizes[
            self.tokenizer.name_or_path
        ]
        self.scorer.eval()
        for param in self.scorer.parameters():
            param.requires_grad = False

        if not cfg.cpu and torch.cuda.is_available():
            if cfg.fp16:
                self.scorer = self.scorer.half()
            elif cfg.bf16:
                self.scorer = self.scorer.bfloat16()
            self.scorer = self.scorer.cuda()

    @property
    def device(self) -> torch.device:
        """Returns the device of the model."""
        return self.scorer.device

    def score(self, hypothesis: str, reference: str, *_, **__) -> float:
        """Calculate the score of the given hypothesis.

        Args:
            hypothesis (str): A hypothesis.
            reference (str): A reference.

        Returns:
            float: The score of the given hypothesis.
        """
        batch = self.tokenizer(
            [reference],
            [hypothesis],
            truncation=True,
            padding=True,
            max_length=self.max_length,
            return_tensors="pt",
        ).to(self.device)
        model_output = self.scorer(**batch)
        return model_output.logits.flatten().tolist()[0]

    def scores(self, hypotheses: list[str], references: list[str], *_, **__) -> Tensor:
        """Calculate the scores of the given hypothesis.

        Args:
            hypotheses (list[str]): N hypotheses.
            references (list[str]): N references.

        Returns:
            Tensor: The N scores of the given hypotheses.
        """

        scores = []
        with timer.measure("score") as t:
            t.set_delta_ncalls(len(hypotheses))
            for i in range(0, len(hypotheses), self.cfg.batch_size):
                batch = self.tokenizer(
                    references[i : i + self.cfg.batch_size],
                    hypotheses[i : i + self.cfg.batch_size],
                    truncation=True,
                    padding=True,
                    max_length=self.max_length,
                    return_tensors="pt",
                ).to(self.device)
                model_output = self.scorer(**batch)
                scores.append(model_output.logits.flatten())
        return torch.cat(scores).view(len(hypotheses))

    def __collate(self, batch_ids_pairs: list[EncodedInputPair]) -> BatchEncoding:
        """
        Prepares a sequence of input id, or a pair of sequences of inputs ids so that it can be used by the model. It
        adds special tokens, truncates sequences if overflowing while taking into account the special tokens and
        manages a moving window (with user defined stride) for overflowing tokens

        Args:
            batch_ids_pairs (list[EncodedInputPair]): List of tokenized input ids or input ids pairs.
        """

        batch = {}
        for first_ids, second_ids in batch_ids_pairs:
            example = self.tokenizer.prepare_for_model(
                first_ids,
                second_ids,
                add_special_tokens=True,
                padding=False,
                truncation=True,
                max_length=self.max_length,
                pad_to_multiple_of=None,
                return_attention_mask=False,
                return_tensors=None,
            )

            for key, value in example.items():
                if key not in batch:
                    batch[key] = []
                batch[key].append(value)

        batch = self.tokenizer.pad(
            batch, padding=True, max_length=self.max_length, return_tensors="pt"
        )
        return batch

    def pairwise_scores(
        self, hypotheses: list[str], references: list[str], *_, **__
    ) -> Tensor:
        """Calculate the pairwise scores.

        Args:
            hypotheses (list[str]): Hypotheses.
            references (list[str]): References.

        Returns:
            Tensor: Score matrix of shape `(H, R)`, where `H` is the number
              of hypotheses and `R` is the number of references.
        """
        scores = []
        hypotheses_ids = [
            self.tokenizer.encode(h, add_special_tokens=False) for h in hypotheses
        ]
        references_ids = [
            self.tokenizer.encode(r, add_special_tokens=False) for r in references
        ]
        pairwise_iter = itertools.product(references_ids, hypotheses_ids)

        while batch := list(itertools.islice(pairwise_iter, self.cfg.batch_size)):
            with timer.measure("score") as t:
                t.set_delta_ncalls(len(batch))
                batch = self.__collate(batch).to(self.device)
                model_output = self.scorer(**batch)
                scores.append(model_output.logits.flatten())
        return torch.cat(scores).view(len(references), len(hypotheses)).transpose(0, 1)

    def corpus_score(
        self, hypotheses: list[str], references: list[str], *_, **__
    ) -> float:
        """Calculate the corpus-level score.

        Args:
            hypotheses (list[str]): Hypotheses.
            references (list[str]): References.

        Returns:
            float: The corpus score.
        """
        scores = []
        for i in range(0, len(hypotheses), self.cfg.batch_size):
            scores.append(
                self.scores(
                    hypotheses[i : i + self.cfg.batch_size],
                    references[i : i + self.cfg.batch_size],
                )
                .float()
                .cpu()
            )
        return torch.cat(scores).mean().item()



from typing import Tuple

import torch
from torch import Generator, Tensor

from mbrs import timer


class Kmeans:
    """k-means clustering implemented in PyTorch.

    Args:
        kmeanspp (bool): Use k-means++ for centroid intialization.
    """

    def __init__(self, kmeanspp: bool = False) -> None:
        self.kmeanspp = kmeanspp

    def assign(self, x: Tensor, centroids: Tensor) -> Tensor:
        """Assigns the nearest neighbor centroid ID.

        Args:
            x (torch.Tensor): Assigned vectors of shape `(n, dim)`.
            centroids (torch.Tensor): Centroids tensor of shape  `(ncentroids, dim)`.

        Returns:
            torch.Tensor: Assigned IDs of shape `(n,)`.
        """
        return torch.cdist(x, centroids, p=2).argmin(dim=-1)

    def init_kmeanspp(self, x: Tensor, rng: Generator, ncentroids: int) -> Tensor:
        """Initializes the centroids via k-means++.

        Args:
            x (Tensor): Input vectors of shape `(n, dim)`.
            rng (Generator): Random number generator.
            ncentroids (int): Number of centroids.

        Returns:
            Tensor: Centroid vectors obtained using k-means++.
        """
        centroids = x[
            torch.randint(x.size(0), size=(1,), generator=rng, device=x.device), :
        ]
        for _ in range(ncentroids - 1):
            # Nc x N
            sqdists = torch.cdist(centroids, x, p=2) ** 2
            neighbor_sqdists = sqdists.min(dim=0).values.float().clamp(min=1e-5)
            weights = neighbor_sqdists / neighbor_sqdists.sum()
            new_centroid = x[torch.multinomial(weights, 1, generator=rng), :]
            centroids = torch.cat([centroids, new_centroid])
        return centroids

    def train(
        self, x: Tensor, ncentroids: int, niter: int = 5, seed: int = 0
    ) -> Tuple[Tensor, Tensor]:
        """Trains k-means.

        Args:
            x (torch.Tensor): Input vectors of shape `(n, dim)`.
            ncentroids (int): Number of centroids.
            niter (int): Number of training iteration.

        Returns:
            Tensor: Centroids tensor of shape `(ncentroids, dim)`.
            Tensor: Assigend IDs of shape `(n,)`.
        """
        if ncentroids == 1:
            with timer.measure("kmeans/iteration"):
                centroids = x.mean(dim=0, keepdim=True)
            return centroids, self.assign(x, centroids)
        elif x.size(0) <= ncentroids:
            return x, torch.arange(x.size(0), device=x.device)

        with timer.measure("kmeans/initialize"):
            rng = torch.Generator(x.device)
            rng = rng.manual_seed(seed)
            if self.kmeanspp:
                centroids = self.init_kmeanspp(x, rng, ncentroids)
            else:
                centroids = x[
                    torch.randperm(x.size(0), generator=rng, device=x.device)[
                        :ncentroids
                    ]
                ]

        assigns = x.new_full((x.size(0),), fill_value=-1)
        for i in range(niter):
            with timer.measure("kmeans/iteration"):
                new_assigns = self.assign(x, centroids)
                if torch.equal(new_assigns, assigns):
                    break
                assigns = new_assigns
                for k in range(ncentroids):
                    if (assigns == k).any():
                        centroids[k] = x[assigns == k].mean(dim=0)
        return centroids, assigns



from typing import Optional, Tuple

import torch
import torch.linalg as LA
from torch import Tensor

from mbrs import timer


class MatrixFactorizationALS:
    """Alternating least squares (ALS) implemented in PyTorch.

    Args:
        regularization_weight (float): Weight of L2 regularization.
        rank (int): Rank of the factarized matrices.
    """

    def __init__(self, regularization_weight: float = 0.1, rank: int = 8) -> None:
        self.regularization_weight = regularization_weight
        self.rank = rank

    def compute_loss(
        self,
        matrix: Tensor,
        x: Tensor,
        y: Tensor,
        observed_mask: Optional[Tensor] = None,
    ) -> float:
        """Compute the objective loss function.

        Args:
            matrix (Tensor): Target matrix of shape `(N, M)`.
            x (Tensor): Left-side low-rank matrix of shape `(N, r)`.
            y (Tensor): Right-side low-rank matrix of shape `(M, r)`.
            observed_mask (Tensor, optional): Valid indices boolean mask of shape `(N, M)`.

        Returns:
            float: Objective loss.
        """
        mse_loss = ((observed_mask * (matrix - x @ (y.T))) ** 2).sum()
        l2_regularization_loss = x.norm() ** 2 + y.norm() ** 2
        loss = mse_loss + self.regularization_weight * l2_regularization_loss
        return loss.item()

    def factorize(
        self,
        matrix: Tensor,
        observed_mask: Optional[Tensor] = None,
        niter: int = 30,
        tolerance: float = 1e-4,
        seed: int = 0,
    ) -> Tuple[Tensor, Tensor]:
        """Factorize the given matrix.

        The input matrix of shape `(N, M)` is decomposed into `X @ Y.T`,
        where `X` and `Y` shape `(N, r)` and `(M, r)`, respectively.

        This implementation does not compute the inverse matrix directly in `X = A^-1 @ b`.
        Instead, `AX = b` is solved.

        Args:
            matrix (Tensor): Input matrix of shape `(N, M)`.
            observed_mask (Tensor, optional): Boolean mask of valid indices of shape `(N, M)`.
            niter (int): The number of alternating steps performed.
            tolerance (float): If the difference between the previous and current loss
              is smaller this value, ALS is regarded as converged.
            seed (int): A seed for the random number generator.

        Returns:
            Tensor: Low-rank matrix `X` of shape `(N, r)`.
            Tensor: Low-rank matrix `Y` of shape `(M, r)`.
        """
        rng = torch.Generator(matrix.device)
        rng = rng.manual_seed(seed)

        N, M = matrix.size()
        # Initialization:
        # Empirically observed the convergence to be much better with the scaled initialization.
        X = (
            torch.rand((N, self.rank), generator=rng, device=matrix.device)
            * (N * self.rank) ** -0.5
        )
        Y = (
            torch.rand((M, self.rank), generator=rng, device=matrix.device)
            * (M * self.rank) ** -0.5
        )
        if observed_mask is None:
            observed_mask = matrix.new_ones((N, M), dtype=torch.bool)

        regularization_term = self.regularization_weight * torch.eye(
            self.rank, device=matrix.device
        )
        prev_loss = float("1e5")
        for _ in range(niter):
            with timer.measure("ALS/iteration"):
                # A: r x r
                # B: r x N
                # Solve Ax = b
                X = LA.solve(
                    Y.T[None, :, :] @ (Y[None, :, :] * observed_mask[:, :, None])
                    + regularization_term,
                    matrix @ Y,
                )
                Y = LA.solve(
                    X.T[None, :, :] @ (X[None, :, :] * observed_mask.T[:, :, None])
                    + regularization_term,
                    matrix.T @ X,
                )
                loss = self.compute_loss(matrix, X, Y, observed_mask=observed_mask)
                if prev_loss - loss <= tolerance:
                    break
                prev_loss = loss
        return X, Y



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

import torch
from torch import Tensor

from mbrs import functional, timer
from mbrs.metrics import MetricCacheable
from mbrs.modules.kmeans import Kmeans
from mbrs.selectors import Selector, SelectorNbest

from . import register
from .mbr import DecoderMBR


@register("centroid_mbr")
class DecoderCentroidMBR(DecoderMBR):
    """Centroid-Based MBR decoder class.

    - Time complexity: O(Nk)
    - Space complexity: O(Nk)

    where k << N.

    References:
        H. Deguchi et al., 2024.
        "Centroid-Based Efficient Minimum Bayes Risk Decoding".
        https://arxiv.org/abs/2402.11197
    """

    def __init__(
        self,
        cfg: DecoderCentroidMBR.Config,
        metric: MetricCacheable,
        selector: Selector = SelectorNbest(SelectorNbest.Config()),
    ) -> None:
        super().__init__(cfg, metric, selector=selector)
        self.kmeans = Kmeans(kmeanspp=self.cfg.kmeanspp)

    cfg: Config

    @dataclass
    class Config(DecoderMBR.Config):
        """Configuration for the decoder.

        - ncentroids (int): Number of centroids.
        - niter (int): Number of k-means iteration
        - count_weight: (bool) Weight the scores with counts.
        - kmeanspp (bool): Initialize the centroids using k-means++.
        - seed (bool): Random seed.
        """

        ncentroids: int = 8
        niter: int = 3
        count_weight: bool = False
        kmeanspp: bool = True
        seed: int = 0

    def decode(
        self,
        hypotheses: list[str],
        references: list[str],
        source: Optional[str] = None,
        nbest: int = 1,
        reference_lprobs: Optional[Tensor] = None,
    ) -> DecoderCentroidMBR.Output:
        """Select the n-best hypotheses based on the strategy.

        Args:
            hypotheses (list[str]): Hypotheses.
            references (list[str]): References.
            source (str, optional): A source.
            nbest (int): Return the n-best hypotheses.
            reference_lprobs (Tensor, optional): Log-probabilities for each reference sample.
              The shape must be `(len(references),)`. See `https://arxiv.org/abs/2311.05263`.

        Returns:
            DecoderCBMBR.Output: The n-best hypotheses.
        """
        assert isinstance(self.metric, MetricCacheable)

        with timer.measure("encode/hypotheses"):
            hypotheses_ir = self.metric.encode(hypotheses)
        if hypotheses == references:
            references_ir = hypotheses_ir
        else:
            with timer.measure("encode/references"):
                references_ir = self.metric.encode(references)
        if source is None:
            source_ir = None
        else:
            with timer.measure("encode/source"):
                source_ir = self.metric.encode([source])
        centroids, assigns = self.kmeans.train(
            references_ir,
            min(self.cfg.ncentroids, len(references)),
            niter=self.cfg.niter,
            seed=self.cfg.seed,
        )

        lprobs = None
        if self.cfg.count_weight:
            centroid_ids, counts_nonzero = assigns.unique(return_counts=True)
            counts = centroid_ids.new_zeros(len(centroids))
            counts[centroid_ids] = counts_nonzero
            lprobs = counts.log()
        elif reference_lprobs is not None:
            reference_lprobs = reference_lprobs.to(centroids)

            # Accumurate the log-probabilities for each centroid by logsumexp.
            lprobs = (
                centroids.new_zeros(len(centroids), dtype=torch.float32)
                .scatter_add(
                    dim=-1,
                    index=assigns.unique(),
                    src=reference_lprobs.softmax(dim=-1, dtype=torch.float32),
                )
                .log()
                .to(centroids)
            )

        with timer.measure("expectation"):
            pairwise_scores = self.metric.pairwise_scores_from_ir(
                hypotheses_ir, centroids, source_ir
            )
            expected_scores = functional.expectation(pairwise_scores, lprobs=lprobs)

        selector_outputs = self.select(
            hypotheses, expected_scores, nbest=nbest, source=source
        )
        return (
            self.Output(
                idx=selector_outputs.idx,
                sentence=selector_outputs.sentence,
                score=selector_outputs.score,
            )
            | selector_outputs
        )



import pytest
import torch

from mbrs.metrics import MetricBLEU, MetricCOMET, MetricCOMETkiwi, Metrics, MetricXCOMET
from mbrs.selectors import SelectorDiverse, SelectorNbest


@pytest.fixture(scope="session")
def metric_comet():
    return MetricCOMET(MetricCOMET.Config())


@pytest.fixture(scope="session")
def metric_cometkiwi():
    return MetricCOMETkiwi(MetricCOMETkiwi.Config())


@pytest.mark.skipif(
    not torch.cuda.is_available(), reason="CUDA is not available on this machine."
)
@pytest.fixture(scope="session")
def metric_xcomet():
    return MetricXCOMET(MetricXCOMET.Config())


@pytest.fixture(
    params=[
        SelectorNbest(SelectorNbest.Config()),
        SelectorDiverse(
            SelectorDiverse.Config(
                diversity_metric=Metrics.bleu,
                diversity_metric_config=MetricBLEU.Config(effective_order=True),
            )
        ),
    ]
)
def selector(request):
    return request.param



from typing import Any

import torch


def to_device(sample: Any, device: torch.device):
    def _to_device(x):
        if torch.is_tensor(x):
            return x.to(device=device, non_blocking=True)
        elif isinstance(x, dict):
            return {key: _to_device(value) for key, value in x.items()}
        elif isinstance(x, list):
            return [_to_device(x) for x in x]
        elif isinstance(x, tuple):
            return tuple(_to_device(x) for x in x)
        elif isinstance(x, set):
            return {_to_device(x) for x in x}
        else:
            return x

    return _to_device(sample)



from typing import Type

import pytest
import torch

from mbrs.metrics import Metric, MetricEnum, Metrics, get_metric

from .diverse import SelectorDiverse


class TestSelectorDiverse:
    @pytest.mark.parametrize("diversity_metric", [Metrics.bleu, Metrics.chrf])
    def test_init(self, diversity_metric: MetricEnum):
        diversity_metric_type: Type[Metric] = get_metric(diversity_metric)
        diversity_metric_config = diversity_metric_type.Config()
        selector = SelectorDiverse(
            SelectorDiverse.Config(
                diversity_metric=diversity_metric,
                diversity_metric_config=diversity_metric_config,
            )
        )
        assert isinstance(selector.diversity_metric, diversity_metric_type)
        assert isinstance(selector.diversity_metric.cfg, diversity_metric_type.Config)

    @pytest.mark.parametrize("diversity_lambda", [0.0, 0.1, 0.5, 1.0, 10.0])
    def test_compute_objective(self, diversity_lambda: float):
        selector = SelectorDiverse(
            SelectorDiverse.Config(diversity_lambda=diversity_lambda)
        )
        expected_scores = torch.arange(5.0)
        hypothesis_dissimilarities = torch.arange(5.0).repeat(5, 1).fill_diagonal_(0.0)
        mask = torch.BoolTensor([True, True, False, True, False])
        obj = selector.compute_objective(
            expected_scores, hypothesis_dissimilarities, mask
        )

        expected_score = (expected_scores[mask].sum() / mask.sum().float()).item()
        _hypothesis_dissimilarities = hypothesis_dissimilarities.clone()
        _hypothesis_dissimilarities[~mask] = 0.0
        _hypothesis_dissimilarities[:, ~mask] = 0.0
        k = mask.sum().item()
        hypothesis_dissimilarity = (
            _hypothesis_dissimilarities.sum() / k / max(k - 1, 1)
        ).item()
        torch.testing.assert_close(obj.expected_score, expected_score)
        torch.testing.assert_close(obj.diversity_score, hypothesis_dissimilarity)
        torch.testing.assert_close(
            obj.score, expected_score + diversity_lambda * hypothesis_dissimilarity
        )

    @pytest.mark.parametrize("maximize", [True, False])
    def test_search_greedy_best_first(self, maximize: bool):
        selector = SelectorDiverse(SelectorDiverse.Config(diversity_lambda=1.1))
        expected_scores = torch.arange(4.0) + 1.0
        hypothesis_dissimilarities = (
            torch.arange(4.0, 0.0, -1.0).repeat(4, 1).fill_diagonal_(0.0)
        )

        mask_patterns = [
            [False, False, False, True],
            [False, False, True, False],
            [False, True, False, False],
            [True, False, False, False],
        ]
        objs = [
            selector.compute_objective(
                expected_scores, hypothesis_dissimilarities, torch.tensor(mask)
            ).score
            for mask in mask_patterns
        ]
        topk_scores, topk_indices = torch.tensor(objs).topk(k=1, largest=maximize)
        selections = selector.search_greedy_best_first(
            expected_scores, hypothesis_dissimilarities, nbest=1, maximize=maximize
        )
        torch.testing.assert_close(
            selections, torch.tensor(mask_patterns[topk_indices.item()])
        )

        mask_patterns = [
            [False, False, True, True],
            [False, True, False, True],
            [True, False, False, True],
            [False, True, True, False],
            [True, False, True, False],
            [True, True, False, False],
        ]
        mask_patterns = [
            mask
            for mask in mask_patterns
            if (torch.tensor(mask) * selections).sum() >= 1
        ]
        objs = [
            selector.compute_objective(
                expected_scores, hypothesis_dissimilarities, torch.tensor(mask)
            ).score
            for mask in mask_patterns
        ]
        topk_scores, topk_indices = torch.tensor(objs).topk(k=1, largest=maximize)
        selections = selector.search_greedy_best_first(
            expected_scores, hypothesis_dissimilarities, nbest=2, maximize=maximize
        )
        torch.testing.assert_close(
            selections, torch.tensor(mask_patterns[topk_indices.item()])
        )

    @pytest.mark.parametrize("maximize", [True, False])
    def test_search_local(self, maximize: bool):
        selector = SelectorDiverse(
            SelectorDiverse.Config(diversity_lambda=1.1, local_search_neighbors=1)
        )
        expected_scores = torch.arange(4.0) + 1.0
        hypothesis_dissimilarities = (
            torch.arange(4.0, 0.0, -1.0).repeat(4, 1).fill_diagonal_(0.0)
        )

        mask_patterns = [
            [False, False, True, True],
            [False, True, False, True],
            [True, False, False, True],
            [False, True, True, False],
            [True, False, True, False],
            [True, True, False, False],
        ]
        solutions = []
        for mask in mask_patterns[:3] if maximize else mask_patterns[3:]:
            selections = selector.search_local(
                expected_scores,
                hypothesis_dissimilarities,
                torch.tensor(mask),
                nbest=2,
                maximize=maximize,
            )
            solutions.append(selections)
        assert torch.unique(torch.stack(solutions), dim=0).size(0) == 1
        objs = [
            selector.compute_objective(
                expected_scores, hypothesis_dissimilarities, torch.tensor(mask)
            ).score
            for mask in mask_patterns
        ]
        topk_scores, topk_indices = torch.tensor(objs).topk(k=1, largest=maximize)
        torch.testing.assert_close(
            torch.unique(torch.stack(solutions), dim=0)[0],
            torch.tensor(mask_patterns[topk_indices.item()]),
        )

    @pytest.mark.parametrize("diversity_metric", [Metrics.bleu, Metrics.chrf])
    @pytest.mark.parametrize("nbest", [1, 3, 5])
    @pytest.mark.parametrize("maximize", [True, False])
    def test_select(self, diversity_metric: MetricEnum, nbest: int, maximize: bool):
        diversity_metric_type: Type[Metric] = get_metric(diversity_metric.value)
        diversity_metric_config = diversity_metric_type.Config()
        selector = SelectorDiverse(
            SelectorDiverse.Config(
                diversity_metric=diversity_metric,
                diversity_metric_config=diversity_metric_config,
                diversity_lambda=0.5,
                local_search_iterations=100,
            )
        )
        scores = torch.Tensor([0.5, 0.9, 0.8, 0.2, 0.4]) * 100.0
        sentences = ["a", "a", "a", "b", "b"]
        output = selector.select(sentences, scores, nbest=nbest, maximize=maximize)
        if maximize:
            if nbest < 5:
                expected_indices = [1, 2, 4, 0, 3]
            else:
                expected_indices = [1, 2, 0, 4, 3]
        else:
            if nbest < 5:
                expected_indices = [3, 4, 0, 2, 1]
            else:
                expected_indices = [3, 4, 0, 2, 1]
        expected_sentences = [sentences[i] for i in expected_indices]
        assert output.idx == expected_indices[:nbest]
        assert output.sentence == expected_sentences[:nbest]



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

import torch
from torch import Tensor

from mbrs import timer
from mbrs.metrics import Metric, Metrics, get_metric
from mbrs.selectors import Selector, register


@register("diverse")
class SelectorDiverse(Selector):
    def __init__(self, cfg: SelectorDiverse.Config) -> None:
        super().__init__(cfg)
        self.diversity_metric: Metric = get_metric(cfg.diversity_metric)(
            cfg.diversity_metric_config
        )

    @dataclass
    class Config(Selector.Config):
        """Configuration for the selector."""

        diversity_metric: Metrics = Metrics.bleu
        diversity_metric_config: Metric.Config | None = None
        diversity_lambda: float = 0.1
        local_search_iterations: int = 20
        local_search_neighbors: int = 1
        seed: int = 0

        def __post_init__(self):
            if self.diversity_metric_config is None:
                self.diversity_metric_config = get_metric(
                    self.diversity_metric
                ).Config()

    cfg: Config

    @dataclass
    class Output(Selector.Output):
        """
        - idx (list[int]): Index numbers of the n-best hypotheses.
        - sentence (list[str]): Sentences of the n-best hypotheses.
        - score (list[float]): Scores of the n-best hypotheses.
        - nbest_objective_score: (float): Objective score for the n-best list.
        - nbest_expected_score (float): Expected score for the n-best list.
        - nbest_diversity_score: (float): Diversity score for the n-best list.
        """

        nbest_objective_score: float
        nbest_expected_score: float
        nbest_diversity_score: float

    @dataclass
    class Objective:
        """
        - score (float): The objective score.
        - expected_score (float): Expected score for the n-best list.
        - diversity_score (float): Diversity score for the n-best list.
        """

        score: float
        expected_score: float
        diversity_score: float

    def compute_objective(
        self, expected_scores: Tensor, hypothesis_dissimilarities: Tensor, mask: Tensor
    ) -> Objective:
        """Compute the objective function.

        Args:
            expected_scores (Tensor): The expected scores for each hypothesis. The shape is `(H,)`.
            hypothesis_dissimilarities (Tensor): The pairwise dissimilarities for all hypotheses. The shape is `(H, H)`.
            mask (Tensor): Boolean tensor of shape `(H,)`. The positions of True elements are calculated
              in the objective and the others are discarded.

        Returns:
            Objective: The objective scores that contain the expected score, diversity score, and the sum of them.
        """
        hypothesis_dissimilarities = hypothesis_dissimilarities.clone().float()
        hypothesis_dissimilarities = hypothesis_dissimilarities.fill_diagonal_(0.0)
        k = mask.sum().item()
        expected_score = expected_scores[mask].float().sum() / k
        hypothesis_dissimilarity = (
            mask.float() @ hypothesis_dissimilarities @ mask.float() / k / max(k - 1, 1)
        )
        objective = (
            expected_score + self.cfg.diversity_lambda * hypothesis_dissimilarity
        ).item()
        return self.Objective(
            objective, expected_score.item(), hypothesis_dissimilarity.item()
        )

    def search_greedy_best_first(
        self,
        expected_scores: Tensor,
        hypothesis_dissimilarities: Tensor,
        nbest: int = 1,
        maximize: bool = True,
    ) -> Tensor:
        """Search the solution by greedy best first search.

        Args:
            expected_scores (Tensor): The expected scores for each hypothesis. The shape is `(H,)`.
            hypothesis_dissimilarities (Tensor): The pairwise dissimilarities for all hypotheses. The shape is `(H, H)`.
            nbest (int): The number of final outputs.
            maximize (bool): Whether maximize the scores or not.

        Returns:
            Tensor: Boolean tensor of shape `(H,)` where True positions indicate that they are selected.
        """
        H = expected_scores.size(0)
        selections = torch.zeros(H, dtype=torch.bool, device=expected_scores.device)
        for k in range(nbest):
            best = float("-inf") if maximize else float("inf")
            best_i = -1
            for i in range(H):
                if selections[i]:
                    continue
                selection_candidate = selections.clone()
                selection_candidate[i] = True
                objective = self.compute_objective(
                    expected_scores, hypothesis_dissimilarities, selection_candidate
                )

                if self.superior(objective.score, best, maximize=maximize):
                    best = objective.score
                    best_i = i
            selections[best_i] = True

        return selections

    def search_local(
        self,
        expected_scores: Tensor,
        hypothesis_dissimilarities: Tensor,
        initial_selections: Tensor,
        nbest: int = 1,
        maximize: bool = True,
    ) -> Tensor:
        """Search the solution by greedy best first search.

        Args:
            expected_scores (Tensor): The expected scores for each hypothesis. The shape is `(H,)`.
            hypothesis_dissimilarities (Tensor): The pairwise dissimilarities for all hypotheses. The shape is `(H, H)`.
            initial_selections (Tensor): Boolean tensor of shape `(H,)` where True positions indicate that they are selected.
            nbest (int): The number of final outputs.
            maximize (bool): Whether maximize the scores or not.

        Returns:
            Tensor: Boolean tensor of shape `(H,)` where True positions indicate that they are selected.
        """
        rng = torch.Generator(device=expected_scores.device).manual_seed(self.cfg.seed)
        H = initial_selections.size(0)
        selections = initial_selections.clone()

        num_neighbors = min(self.cfg.local_search_neighbors, nbest)

        for i in range(self.cfg.local_search_iterations):
            prev_selections = selections.clone()
            selection_indices = selections.nonzero(as_tuple=True)[0]
            removed_candidates = torch.randperm(
                nbest, generator=rng, device=rng.device
            )[:num_neighbors]

            for k in range(num_neighbors):
                selections[selection_indices[removed_candidates[k]]] = False

            for k in range(num_neighbors):
                best = float("-inf") if maximize else float("inf")
                best_i = -1
                for i in range(H):
                    if selections[i]:
                        continue
                    selection_candidate = selections.clone()
                    selection_candidate[i] = True
                    objective = self.compute_objective(
                        expected_scores, hypothesis_dissimilarities, selection_candidate
                    )
                    if self.superior(objective.score, best, maximize=maximize):
                        best = objective.score
                        best_i = i
                selections[best_i] = True

            prev_objective = self.compute_objective(
                expected_scores, hypothesis_dissimilarities, prev_selections
            )
            new_objective = self.compute_objective(
                expected_scores, hypothesis_dissimilarities, selections
            )
            if self.superior(
                prev_objective.score, new_objective.score, maximize=maximize
            ):
                selections = prev_selections

        return selections

    def select(
        self,
        hypotheses: list[str],
        expected_scores: Tensor,
        nbest: int = 1,
        source: Optional[str] = None,
        maximize: bool = True,
        *args,
        **kwargs,
    ) -> SelectorDiverse.Output:
        """Select the final output list.

        Args:
            hypotheses (list[str]): Hypotheses.
            expected_scores (Tensor): The expected scores for each hypothesis.
            nbest (int): Return the n-best hypotheses based on the selection rule.
            source (str, optional): A source.
            maximize (bool): Whether maximize the scores or not.

        Returns:
            Selector.Output: Selected hypotheses.
        """
        nbest = min(len(hypotheses), nbest)
        with timer.measure("dissimilarity_calculation"):
            hypothesis_dissimilarities = self.diversity_metric.pairwise_scores(
                hypotheses,
                hypotheses,
                source=source,
            ).to(expected_scores)
            if maximize:
                hypothesis_dissimilarities *= -1
        with timer.measure("search/greedy_best_first"):
            selections = self.search_greedy_best_first(
                expected_scores,
                hypothesis_dissimilarities,
                nbest=nbest,
                maximize=maximize,
            )
        with timer.measure("search/local"):
            selections = self.search_local(
                expected_scores,
                hypothesis_dissimilarities,
                selections,
                nbest=nbest,
                maximize=maximize,
            )
        objective = self.compute_objective(
            expected_scores, hypothesis_dissimilarities, selections
        )
        topk_scores, topk_order = self.topk(
            expected_scores[selections].float(), k=nbest, maximize=maximize
        )
        selected_idx_set = [
            i
            for i, selection in zip(range(len(hypotheses)), selections.tolist())
            if selection
        ]
        topk_indices = [selected_idx_set[i] for i in topk_order]
        return self.Output(
            idx=topk_indices,
            sentence=[hypotheses[idx] for idx in topk_indices],
            score=topk_scores,
            nbest_objective_score=objective.score,
            nbest_expected_score=objective.expected_score,
            nbest_diversity_score=objective.diversity_score,
        )



from typing import List, Tuple, Union

import torch

from .als import MatrixFactorizationALS

N = 64
M = 128
r = 8


def is_equal_shape(
    a: Union[torch.Tensor, torch.Size],
    b: Union[torch.Tensor, torch.Size, List[int], Tuple[int]],
) -> bool:
    """Returns whether a and b have the same shape.

    Args:
        a (Union[torch.Tensor, torch.Size]): An input tensor.
        b (Union[torch.Tensor, torch.Size, List[int], Tuple[int]):
          An input tensor compared the shape with a.

    Returns:
        bool: Whether a and b have the same shape.

    Raises:
        NotImplementedError: When an unsupported type object is given.
    """
    if isinstance(a, torch.Tensor):
        a_shape = a.shape
    elif isinstance(a, torch.Size):
        a_shape = a
    else:
        raise NotImplementedError(f"Type of `a` (`{type(a)}`) is not supported.")

    if isinstance(b, torch.Tensor):
        b_shape = b.shape
    elif isinstance(b, torch.Size):
        b_shape = b
    elif isinstance(b, (list, tuple)):
        b_shape = torch.Size(b)
    else:
        raise NotImplementedError(f"Type of `b` (`{type(b)}`) is not supported.")
    return a_shape == b_shape


class TestMatrixFactorizationALS:
    def test_factorize(self):
        als = MatrixFactorizationALS(rank=r)

        torch.manual_seed(0)
        matrix = torch.rand(N, M)

        X, Y = als.factorize(matrix)
        assert is_equal_shape(X, [N, r])
        assert is_equal_shape(Y, [M, r])



from __future__ import annotations

from typing import Optional

from torch import Tensor

from mbrs.selectors import Selector, register


@register("nbest")
class SelectorNbest(Selector):
    def select(
        self,
        hypotheses: list[str],
        expected_scores: Tensor,
        nbest: int = 1,
        source: Optional[str] = None,
        maximize: bool = True,
        *args,
        **kwargs,
    ) -> SelectorNbest.Output:
        """Select the final output list.

        Args:
            hypotheses (list[str]): Hypotheses.
            expected_scores (Tensor): The expected scores for each hypothesis.
            nbest (int): Return the n-best hypotheses based on the selection rule.
            source (str, optional): A source.
            maximize (bool): Whether maximize the scores or not.

        Returns:
            Selector.Output: Selected hypotheses.
        """
        nbest = min(len(hypotheses), nbest)
        topk_scores, topk_indices = self.topk(
            expected_scores, k=nbest, maximize=maximize
        )
        return self.Output(
            idx=topk_indices,
            sentence=[hypotheses[idx] for idx in topk_indices],
            score=topk_scores,
        )



import enum

from mbrs import registry

from .base import DecoderBase, DecoderReferenceBased, DecoderReferenceless

register, get_decoder = registry.setup("decoder")

from .aggregate_mbr import DecoderAggregateMBR
from .centroid_mbr import DecoderCentroidMBR
from .mbr import DecoderMBR
from .probabilistic_mbr import DecoderProbabilisticMBR
from .pruning_mbr import DecoderPruningMBR
from .rerank import DecoderRerank

__all__ = [
    "DecoderBase",
    "DecoderReferenceBased",
    "DecoderReferenceless",
    "DecoderMBR",
    "DecoderAggregateMBR",
    "DecoderCentroidMBR",
    "DecoderProbabilisticMBR",
    "DecoderPruningMBR",
    "DecoderRerank",
]

Decoders = enum.Enum("Decoders", registry.get_registry("decoder"))



from __future__ import annotations

import concurrent.futures
import itertools
import math
from collections import Counter, defaultdict
from dataclasses import dataclass
from typing import Optional

import torch
from sacrebleu.metrics.chrf import CHRF
from sacrebleu.metrics.helpers import extract_all_char_ngrams, extract_word_ngrams
from torch import Tensor

from mbrs import timer

from . import Metric, MetricAggregatable, register


@register("chrf")
class MetricChrF(MetricAggregatable):
    """ChrF metric class."""

    @dataclass
    class Config(Metric.Config):
        """ChrF metric configuration.

        - char_order (int): Character n-gram order.
        - word_order (int): Word n-gram order. If equals to 2, the metric is referred to as chrF++.
        - beta (int): Determine the importance of recall w.r.t precision.
        - lowercase (bool): Enable case-insensitivity.
        - whitespace (bool): If `True`, include whitespaces when extracting character n-grams.
        - eps_smoothing (bool): If `True`, applies epsilon smoothing similar to reference chrF++.py, NLTK and Moses implementations.
            Otherwise, it takes into account effective match order similar to sacreBLEU < 2.0.0.
        - num_workers (int): Number of workers for multiprocessing.
        """

        char_order: int = 6
        word_order: int = 0
        beta: int = 2
        lowercase: bool = False
        whitespace: bool = False
        eps_smoothing: bool = False
        num_workers: int = 8

    cfg: Config

    @dataclass
    class AggregatedReference:
        """Aggregated reference representation.

        - ngrams (list[Counter]]): Bags of n-grams for each order.
        """

        ngrams: list[Counter]

    def __init__(self, cfg: MetricChrF.Config):
        super().__init__(cfg)
        self.scorer = CHRF(
            char_order=cfg.char_order,
            word_order=cfg.word_order,
            beta=cfg.beta,
            lowercase=cfg.lowercase,
            whitespace=cfg.whitespace,
            eps_smoothing=cfg.eps_smoothing,
        )

    def score(self, hypothesis: str, reference: str, *_, **__) -> float:
        """Calculate the score of the given hypothesis.

        Args:
            hypothesis (str): Hypothesis.
            reference (str): Reference.

        Returns:
            float: The score of the given hypothesis.
        """
        return self.scorer.sentence_score(hypothesis, [reference]).score

    def scores(self, hypotheses: list[str], references: list[str], *_, **__) -> Tensor:
        """Calculate the scores of the given hypotheses.

        Args:
            hypotheses (list[str]): N hypotheses.
            references (list[str]): N references.

        Returns:
            Tensor: The N scores of the given hypotheses.
        """
        with concurrent.futures.ProcessPoolExecutor(
            max_workers=self.cfg.num_workers,
        ) as executor:
            with timer.measure("score") as t:
                t.set_delta_ncalls(len(hypotheses))
                return Tensor(
                    list(
                        executor.map(
                            self.score,
                            hypotheses,
                            references,
                            chunksize=math.ceil(len(hypotheses) / self.cfg.num_workers),
                        )
                    )
                )

    def pairwise_scores(
        self, hypotheses: list[str], references: list[str], *_, **__
    ) -> Tensor:
        """Calculate the pairwise scores.

        Args:
            hypotheses (list[str]): Hypotheses.
            references (list[str]): References.

        Returns:
            Tensor: Score matrix of shape `(H, R)`, where `H` is the number
              of hypotheses and `R` is the number of references.
        """
        with concurrent.futures.ProcessPoolExecutor(
            max_workers=self.cfg.num_workers
        ) as executor:
            with timer.measure("score") as t:
                t.set_delta_ncalls(len(hypotheses) * len(references))

                return Tensor(
                    list(
                        executor.map(
                            self.score,
                            *zip(*itertools.product(hypotheses, references)),
                            chunksize=len(hypotheses),
                        )
                    )
                ).view(len(hypotheses), len(references))

    def corpus_score(
        self, hypotheses: list[str], references: list[str], *_, **__
    ) -> float:
        """Calculate the corpus-level score.

        Args:
            hypotheses (list[str]): Hypotheses.
            references (list[str]): References.

        Returns:
            float: The corpus score.
        """
        return self.scorer.corpus_score(hypotheses, [references]).score

    def _aggregate_references(
        self, references: list[str], reference_lprobs: Optional[Tensor] = None
    ) -> AggregatedReference:
        """Aggregate references.

        Args:
            references (list[str]): References.
            reference_lprobs (Tensor, optional): Log-probabilities for each reference sample.
              The shape must be `(len(references),)`. See `https://arxiv.org/abs/2311.05263`.

        Returns:
            MetricChrF.AggregatedReference: Aggregated reference representation.
        """
        num_references = len(references)
        reference_ngrams: list[list[Counter[str]]] = self.scorer._cache_references(
            [[ref] for ref in references]
        )[0]["ref_ngrams"]

        if reference_lprobs is not None:
            lprobs = reference_lprobs.log_softmax(dim=-1).tolist()
        else:
            lprobs = [-math.log(num_references)] * num_references

        acc_ngrams: defaultdict[int, Counter[str]] = defaultdict(Counter)
        for i, ngrams in enumerate(reference_ngrams):
            for order, ngram_counts in enumerate(ngrams):
                for ngram in ngram_counts:
                    # Note: Counter has float values.
                    ngram_counts[ngram] = math.exp(
                        math.log(ngram_counts[ngram]) + lprobs[i]
                    )
                acc_ngrams[order] += ngram_counts

        return self.AggregatedReference(
            [acc_ngrams[order] for order in range(len(acc_ngrams))]
        )

    def expected_scores_reference_aggregation(
        self,
        hypotheses: list[str],
        references: list[str],
        source: Optional[str] = None,
        reference_lprobs: Optional[Tensor] = None,
    ) -> Tensor:
        """Calculate the expected scores for each hypothesis.

        Args:
            hypotheses (list[str]): Hypotheses.
            references (list[str]): References.
            source (str, optional): A source.
            reference_lprobs (Tensor, optional): Log-probabilities for each reference sample.
              The shape must be `(len(references),)`. See `https://arxiv.org/abs/2311.05263`.

        Returns:
            Tensor: The expected scores for each hypothesis.
        """
        with timer.measure("aggregate/references"):
            aggregated_reference = self._aggregate_references(
                references, reference_lprobs=reference_lprobs
            )

        expected_scores = torch.zeros((len(hypotheses),))
        for i, hypothesis in enumerate(hypotheses):
            with timer.measure("expectation"):
                hypothesis = self.scorer._preprocess_segment(hypothesis)

                # Extract character n-grams
                all_hyp_ngrams = extract_all_char_ngrams(
                    hypothesis, self.scorer.char_order, self.scorer.whitespace
                )

                # Check chrF+ mode to see if we'll add word n-grams as well
                if self.scorer.word_order > 0:
                    # Primitive tokenization: separate out punctuations
                    hwords = self.scorer._remove_punctuation(hypothesis)
                    _range = range(1, self.scorer.word_order + 1)
                    all_hyp_ngrams.extend(
                        [extract_word_ngrams(hwords, n) for n in _range]
                    )

                stats = []
                # Traverse all orders
                for h, r in zip(all_hyp_ngrams, aggregated_reference.ngrams):
                    stats.extend(self.scorer._get_match_statistics(h, r))
                f_score = self.scorer._compute_f_score(stats)
                expected_scores[i] = f_score

        return expected_scores



from __future__ import annotations

import abc
from dataclasses import dataclass
from typing import Optional

import torch
from torch import Tensor

from mbrs import functional, timer


class MetricBase(abc.ABC):
    """Base metric class."""

    def __init__(self, cfg: MetricBase.Config):
        self.cfg = cfg

    HIGHER_IS_BETTER: bool = True

    @dataclass
    class Config: ...

    @property
    def device(self) -> torch.device:
        """Returns the device of the metric object."""
        return torch.device("cpu")


class Metric(MetricBase, metaclass=abc.ABCMeta):
    """Base metric class."""

    @abc.abstractmethod
    def score(
        self, hypothesis: str, reference: str, source: Optional[str] = None
    ) -> float:
        """Calculate the score of the given hypothesis.

        Args:
            hypothesis (str): A hypothesis.
            reference (str): A reference.
            source (str, optional): A source.

        Returns:
            float: The score of the given hypothesis.
        """

    def scores(
        self,
        hypotheses: list[str],
        references: list[str],
        sources: Optional[list[str]] = None,
    ) -> Tensor:
        """Calculate the scores of the given hypotheses.

        Args:
            hypotheses (list[str]): N hypotheses.
            references (list[str]): N references.
            sources (list[str], optional): N sources.

        Returns:
            Tensor: The N scores of the given hypotheses.
        """
        with timer.measure("score") as t:
            t.set_delta_ncalls(len(hypotheses))
            if sources is None:
                return Tensor(
                    [self.score(hyp, ref) for hyp, ref in zip(hypotheses, references)]
                )
            else:
                return Tensor(
                    [
                        self.score(hyp, ref, src)
                        for hyp, ref, src in zip(hypotheses, references, sources)
                    ]
                )

    def pairwise_scores(
        self, hypotheses: list[str], references: list[str], source: Optional[str] = None
    ) -> Tensor:
        """Calculate the pairwise scores.

        Args:
            hypotheses (list[str]): Hypotheses.
            references (list[str]): References.
            source (str, optional): A source.

        Returns:
            Tensor: Score matrix of shape `(H, R)`, where `H` is the number
              of hypotheses and `R` is the number of references.
        """
        with timer.measure("score") as t:
            t.set_delta_ncalls(len(hypotheses) * len(references))
            return Tensor(
                [
                    [self.score(hyp, ref, source) for ref in references]
                    for hyp in hypotheses
                ]
            )

    def expected_scores(
        self,
        hypotheses: list[str],
        references: list[str],
        source: Optional[str] = None,
        reference_lprobs: Optional[Tensor] = None,
    ) -> Tensor:
        """Calculate the expected scores for each hypothesis.

        Args:
            hypotheses (list[str]): Hypotheses.
            references (list[str]): References.
            source (str, optional): A source.
            reference_lprobs (Tensor, optional): Log-probabilities for each reference sample.
              The shape must be `(len(references),)`. See `https://arxiv.org/abs/2311.05263`.

        Returns:
            Tensor: The expected scores for each hypothesis.
        """
        with timer.measure("expectation"):
            return functional.expectation(
                self.pairwise_scores(hypotheses, references, source),
                lprobs=reference_lprobs,
            )

    def corpus_score(
        self,
        hypotheses: list[str],
        references: list[str],
        sources: Optional[list[str]] = None,
    ) -> float:
        """Calculate the corpus-level score.

        Args:
            hypotheses (list[str]): Hypotheses.
            references (list[str]): References.
            sources (list[str], optional): Sources.

        Returns:
            float: The corpus score.
        """
        if sources is None:
            return sum(
                [self.score(hyp, ref) for hyp, ref in zip(hypotheses, references)]
            ) / len(hypotheses)
        else:
            return sum(
                [
                    self.score(hyp, ref, src)
                    for hyp, ref, src in zip(hypotheses, references, sources)
                ]
            ) / len(hypotheses)


class MetricAggregatable(Metric, metaclass=abc.ABCMeta):
    """Base class for aggregatable metrics.

    This class supports reference aggregation."""

    @abc.abstractmethod
    def expected_scores_reference_aggregation(
        self,
        hypotheses: list[str],
        references: list[str],
        source: Optional[str] = None,
        reference_lprobs: Optional[Tensor] = None,
    ) -> Tensor:
        """Calculate the expected scores for each hypothesis.

        Args:
            hypotheses (list[str]): Hypotheses.
            references (list[str]): References.
            source (str, optional): A source.
            reference_lprobs (Tensor, optional): Log-probabilities for each reference sample.
              The shape must be `(len(references),)`. See `https://arxiv.org/abs/2311.05263`.

        Returns:
            Tensor: The expected scores for each hypothesis.
        """


class MetricCacheable(MetricAggregatable, metaclass=abc.ABCMeta):
    """Base class for cacheable metrics.

    This class supports to cache intermediate representations of the encoder."""

    @property
    @abc.abstractmethod
    def embed_dim(self) -> int:
        """Return the size of embedding dimension."""

    @abc.abstractmethod
    def encode(self, sentences: list[str]) -> Tensor:
        """Encode the given sentences into their intermediate representations.

        Args:
            sentences (list[str]): Input sentences.

        Returns:
            Tensor: Intermediate representations of shape `(N, D)` where `N` is the
              number of hypotheses and `D` is a size of the embedding dimension.
        """

    @abc.abstractmethod
    def out_proj(
        self,
        hypotheses_ir: Tensor,
        references_ir: Tensor,
        sources_ir: Optional[Tensor] = None,
    ) -> Tensor:
        """Forward the output projection layer.

        Args:
            hypotheses_ir (Tensor): Intermediate representations of hypotheses.
            references_ir (Tensor): Intermediate representations of references.
            sources_ir (Tensor, optional): Intermediate representations of sources.

        Returns:
            Tensor: N scores.
        """

    def score(
        self,
        hypothesis: str,
        reference: str,
        source: Optional[str] = None,
    ) -> float:
        """Calculate the score of the given hypothesis.

        Args:
            hypothesis (str): A hypothesis.
            reference (str): A reference.
            source (str, optional): A source.

        Returns:
            float: The score of the given hypothesis.
        """
        return self.out_proj(
            self.encode([hypothesis]),
            self.encode([reference]),
            self.encode([source]) if source is not None else None,
        ).item()

    def scores_from_ir(
        self,
        hypotheses_ir: Tensor,
        references_ir: Tensor,
        sources_ir: Optional[Tensor] = None,
    ) -> Tensor:
        """Calculate the scores of the given hypotheses from the intermediate representations.

        Args:
            hypotheses_ir (Tensor): N hypotheses.
            references_ir (Tensor): N references.
            sources_ir (Tensor, optional): N sources.

        Returns:
            Tensor: The N scores of the given hypotheses.
        """
        H = len(hypotheses_ir)
        with timer.measure("score") as t:
            t.set_delta_ncalls(H)
            if sources_ir is None:
                return self.out_proj(hypotheses_ir, references_ir)
            else:
                return self.out_proj(hypotheses_ir, references_ir, sources_ir)

    def scores(
        self,
        hypotheses: list[str],
        references: list[str],
        sources: Optional[list[str]] = None,
    ) -> Tensor:
        """Calculate the scores of the given hypotheses.

        Args:
            hypotheses (list[str]): N hypotheses.
            references (list[str]): N references.
            source (list[str], optional): N sources.

        Returns:
            Tensor: The N scores of the given hypotheses.
        """

        return self.scores_from_ir(
            self.encode(hypotheses),
            self.encode(references),
            self.encode(sources) if sources is not None else None,
        )

    def pairwise_scores_from_ir(
        self,
        hypotheses_ir: Tensor,
        references_ir: Tensor,
        source_ir: Optional[Tensor] = None,
    ) -> Tensor:
        """Calculate the pairwise scores from the intermediate representations.

        Args:
            hypotheses_ir (Tensor): Hypotheses.
            references_ir (Tensor): References.
            source_ir (Tensor, optional): A source.

        Returns:
            Tensor: Score matrix of shape `(H, R)`, where `H` is the number
              of hypotheses and `R` is the number of references.
        """
        H, D = hypotheses_ir.size()
        R, _ = references_ir.size()
        if source_ir is not None:
            source_ir = source_ir.repeat(H, 1)

        scores = []
        for i in range(R):
            with timer.measure("score") as t:
                t.set_delta_ncalls(H)
                scores.append(
                    self.out_proj(
                        hypotheses_ir, references_ir[i, :].repeat(H, 1), source_ir
                    )[:, None]
                )
        return torch.cat(scores, dim=-1)

    def pairwise_scores(
        self, hypotheses: list[str], references: list[str], source: Optional[str] = None
    ) -> Tensor:
        """Calculate the pairwise scores.

        Args:
            hypotheses (list[str]): Hypotheses.
            references (list[str]): References.
            source (str, optional): A source.

        Returns:
            Tensor: Score matrix of shape `(H, R)`, where `H` is the number
              of hypotheses and `R` is the number of references.
        """
        with timer.measure("encode/hypotheses"):
            hypotheses_ir = self.encode(hypotheses)
        if hypotheses == references:
            references_ir = hypotheses_ir
        else:
            with timer.measure("encode/references"):
                references_ir = self.encode(references)
        if source is None:
            source_ir = None
        else:
            with timer.measure("encode/source"):
                source_ir = self.encode([source])
        return self.pairwise_scores_from_ir(hypotheses_ir, references_ir, source_ir)

    def expected_scores_reference_aggregation(
        self,
        hypotheses: list[str],
        references: list[str],
        source: Optional[str] = None,
        reference_lprobs: Optional[Tensor] = None,
    ) -> Tensor:
        """Calculate the expected scores for each hypothesis.

        Args:
            hypotheses (list[str]): Hypotheses.
            references (list[str]): References.
            source (str, optional): A source.
            reference_lprobs (Tensor, optional): Log-probabilities for each reference sample.
              The shape must be `(len(references),)`. See `https://arxiv.org/abs/2311.05263`.

        Returns:
            Tensor: The expected scores for each hypothesis.
        """
        with timer.measure("encode/hypotheses"):
            hypotheses_ir = self.encode(hypotheses)
        if hypotheses == references:
            references_ir = hypotheses_ir
        else:
            with timer.measure("encode/references"):
                references_ir = self.encode(references)
        if source is None:
            source_ir = None
        else:
            with timer.measure("encode/source"):
                source_ir = self.encode([source])

        with timer.measure("aggregate/references"):
            if reference_lprobs is not None:
                aggregated_reference_ir = (
                    references_ir
                    * reference_lprobs.to(references_ir)
                    .softmax(dim=-1, dtype=torch.float32)
                    .to(references_ir)[:, None]
                ).sum(dim=0, keepdim=True)
            else:
                aggregated_reference_ir = references_ir.mean(dim=0, keepdim=True)

        with timer.measure("expectation"):
            return self.pairwise_scores_from_ir(
                hypotheses_ir, aggregated_reference_ir, source_ir=source_ir
            ).mean(dim=-1)


class MetricReferenceless(MetricBase, metaclass=abc.ABCMeta):
    """Base class for reference-less metrics like quality estimation."""

    @abc.abstractmethod
    def score(self, hypothesis: str, source: str) -> float:
        """Calculate the score of the given hypothesis.

        Args:
            hypothesis (str): A hypothesis.
            source (str): A source.

        Returns:
            float: The score of the given hypothesis.
        """

    def scores(self, hypotheses: list[str], sources: list[str]) -> Tensor:
        """Calculate the scores of hypotheses.

        Args:
            hypotheses (list[str]): N hypotheses.
            sources (list[str]): N sources.

        Returns:
            Tensor: The scores of hypotheses.
        """
        return Tensor([self.score(hyp, src) for hyp, src in zip(hypotheses, sources)])

    def corpus_score(self, hypotheses: list[str], sources: list[str]) -> float:
        """Calculate the corpus-level score.

        Args:
            hypotheses (list[str]): Hypotheses.
            sources (list[str]): Sources.

        Returns:
            float: The corpus score.
        """
        return sum(self.scores(hypotheses, sources=sources).cpu().float().item()) / len(
            hypotheses
        )



import pytest
import torch

from mbrs.metrics import MetricChrF, MetricCOMET
from mbrs.selectors import Selector

from .probabilistic_mbr import DecoderProbabilisticMBR

SOURCE = [
    "これはテストです",
    "これはテストです",
    "これはテストです",
    "これはテストです",
]
HYPOTHESES = [
    ["another test", "this is a test", "this is a fest", "x", "this is test"],
    ["another test", "this is a fest", "this is a test"],
    ["this is a test"],
    ["Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;"],
]
REFERENCES = [
    ["another test", "this is a test", "this is a fest", "x", "this is test"],
    ["this is a test", "ref", "these are tests", "this is the test"],
    ["this is a test"],
    ["producţia de zahăr brut se exprimă în zahăr alb;"],
]

BEST_INDICES = [1, 2, 0, 0]
BEST_SENTENCES = [
    "this is a test",
    "this is a test",
    "this is a test",
    "Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;",
]

NITER = 30
FACTOR = 1.25
RANK = 2


class TestDecoderProbabilisticMBR:
    def test_decode_chrf(self):
        metric = MetricChrF(MetricChrF.Config())
        decoder = DecoderProbabilisticMBR(
            DecoderProbabilisticMBR.Config(
                reduction_factor=FACTOR, rank=RANK, niter=NITER
            ),
            metric,
        )
        for i, (hyps, refs) in enumerate(zip(HYPOTHESES, REFERENCES)):
            output = decoder.decode(hyps, refs, SOURCE[i], nbest=1)
            assert output.idx[0] == BEST_INDICES[i]
            assert output.sentence[0] == BEST_SENTENCES[i]

    def test_decode_comet(self, metric_comet: MetricCOMET):
        decoder = DecoderProbabilisticMBR(
            DecoderProbabilisticMBR.Config(
                reduction_factor=FACTOR, rank=RANK, niter=NITER
            ),
            metric_comet,
        )
        for i, (hyps, refs) in enumerate(zip(HYPOTHESES, REFERENCES)):
            output = decoder.decode(hyps, refs, SOURCE[i], nbest=1)
            assert output.idx[0] == BEST_INDICES[i]
            assert output.sentence[0] == BEST_SENTENCES[i]

            output = decoder.decode(
                hyps,
                refs,
                SOURCE[i],
                nbest=1,
                reference_lprobs=torch.Tensor([-2.000]).repeat(len(refs)),
            )
            assert output.idx[0] == BEST_INDICES[i]
            assert output.sentence[0] == BEST_SENTENCES[i]

    @pytest.mark.parametrize("nbest", [1, 2])
    def test_decode_selector(self, nbest: int, selector: Selector):
        metric = MetricChrF(MetricChrF.Config())
        decoder = DecoderProbabilisticMBR(
            DecoderProbabilisticMBR.Config(
                reduction_factor=FACTOR, rank=RANK, niter=NITER
            ),
            metric,
            selector=selector,
        )
        for i, (hyps, refs) in enumerate(zip(HYPOTHESES, REFERENCES)):
            output = decoder.decode(hyps, refs, SOURCE[i], nbest=nbest)
            assert len(output.sentence) == min(nbest, len(hyps))
            assert len(output.score) == min(nbest, len(hyps))

            output = decoder.decode(
                hyps,
                refs,
                SOURCE[i],
                nbest=nbest,
                reference_lprobs=torch.Tensor([-2.000]).repeat(len(refs)),
            )
            assert len(output.sentence) == min(nbest, len(hyps))
            assert len(output.score) == min(nbest, len(hyps))
